from collections import defaultdict

from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType

GATEWAY_HELLO_MESSAGES = [GatewayMessageType.HELLO, BloxrouteMessageType.ACK]

GATEWAY_BLOCKS_SEEN_EXPIRATION_TIME_S = 60 * 60 * 24

# Delay for blockchain sync message request before broadcasting to everyone
# This constants is currently unused
BLOCKCHAIN_SYNC_BROADCAST_DELAY_S = 5
BLOCKCHAIN_PING_INTERVAL_S = 15

BLOCK_RECOVERY_RECOVERY_INTERVAL_S = [0.1, 0.5, 1, 2, 5]
# BLOCK_RECOVERY_MAX_RETRY_ATTEMPTS = len(BLOCK_RECOVERY_RECOVERY_INTERVAL_S)
BLOCK_RECOVERY_MAX_RETRY_ATTEMPTS = 1  # for now, since longer retries aren't really worth it
BLOCK_RECOVERY_MAX_QUEUE_TIME = 15  # slightly more than sum(BLOCK_RECOVERY_RECOVERY_INTERVAL_S)
CHECK_MEMORY_THRESHOLD_INTERVAL_S = 60 * 60
CHECK_MEMORY_THRESHOLD_LIMIT = 4 * 1024 * 1024 * 1024

# enum for setting Gateway neutrality assertion policy for releasing encryption keys
class NeutralityPolicy:
    RECEIPT_COUNT = 1
    RECEIPT_PERCENT = 2
    RECEIPT_COUNT_AND_PERCENT = 3

    RELEASE_IMMEDIATELY = 99


# duration to wait for block receipts until timeout
NEUTRALITY_BROADCAST_BLOCK_TIMEOUT_S = 3

NEUTRALITY_POLICY = NeutralityPolicy.RECEIPT_PERCENT
NEUTRALITY_EXPECTED_RECEIPT_COUNT = 1
NEUTRALITY_EXPECTED_RECEIPT_PERCENT = 50

# Max duration to wait before releasing a block, even if blockchain node has not indicated receipt of
# previous block in chain. This value can be set to 0 if a blockchain node implementation is capable of
# immediately taking block messages without validating previous block.

MAX_INTERVAL_BETWEEN_BLOCKS_S = 0.6
NODE_READINESS_FOR_BLOCKS_CHECK_INTERVAL_S = 5
MAX_BLOCK_CACHE_TIME_S = 20 * 60
MAX_BLOCK_BACKLOG_TO_PUBLISH = 10

GATEWAY_TRANSACTION_STATS_INTERVAL_S = 1 * 60
GATEWAY_TRANSACTION_STATS_LOOKBACK = 1
GATEWAY_BDN_PERFORMANCE_STATS_INTERVAL_S = 15 * 60
GATEWAY_BDN_PERFORMANCE_STATS_LOOKBACK = 1
GATEWAY_TRANSACTION_FEED_STATS_INTERVAL_S = 5 * 60
GATEWAY_TRANSACTION_FEED_STATS_LOOKBACK = 1

MIN_PEER_RELAYS_BY_COUNTRY = defaultdict(lambda: 1)
MAX_PEER_RELAYS_COUNT = 2

BLOCKCHAIN_SOCKET_SEND_BUFFER_SIZE = 1 * 1024 * 1024

COOKIE_FILE_PATH_TEMPLATE = ".gateway_cookies/.cookie.blxrbdn-gw-{}"

SEND_REQUEST_RELAY_PEERS_MAX_NUM_OF_CALLS = 10
SEND_REQUEST_GATEWAY_PEERS_MAX_NUM_OF_CALLS = 5

GATEWAY_PEER_NAME = "bloXroute Gateway"
CONFIG_UPDATE_INTERVAL_S = 360
CONFIG_FILE_NAME = "config.default.json"
CONFIG_OVERRIDE_NAME = "config.override.json"

BLOCK_HANDLING_TIME_EXPIRATION_TIME_S = 300

INITIAL_LIVELINESS_CHECK_S = 100
DEFAULT_STAY_ALIVE_DURATION_S = 30 * 60
CHECK_BLOCKCHAIN_CONNECTION_FIRMLY_ESTABLISHED = 5 * 60

ACTIVE_BLOCKCHAIN_PEERS_LIVELINESS_CHECK_S = 30 * 60

DEFAULT_BLOCKCHAIN_MESSAGE_TTL_S = 10
DEFAULT_REMOTE_BLOCKCHAIN_MESSAGE_TTL_S = 10

BLOCK_CLEANUP_NODE_BLOCK_LIST_POLL_INTERVAL_S = 60
TRACKED_BLOCK_CLEANUP_INTERVAL_S = 0

# ignore last confirmed block and request block confirmation since last tracked block instead
BLOCK_CLEANUP_REQUEST_EXPECTED_ADDITIONAL_TRACKED_BLOCKS = 1

REMOTE_BLOCKCHAIN_MAX_CONNECT_RETRIES = 10
REMOTE_BLOCKCHAIN_SDN_CONTACT_RETRY_SECONDS = 30

MULTIPLE_BLOCKCHAIN_NODE_MAX_RETRY_S = 300

BLOCK_CONFIRMATION_EXPIRE_TIME_S = 60 * 60

LOGGING_LIMIT_ITEM_COUNT = 10

RELAY_CONNECTION_REEVALUATION_INTERVAL_S = 4 * 60 * 60

BLOCK_QUEUE_LENGTH_LIMIT = 128
MSG_PROXY_REQUESTER_QUEUE_LIMIT = 15

LOCALHOST = "127.0.0.1"

TRACKED_BLOCK_MAX_HASH_LOOKUP = 100

BDN_TX_PROCESSING_TIME_WARNING_THRESHOLD_S = 0.05
BLOCKCHAIN_TX_PROCESSING_TIME_WARNING_THRESHOLD_S = 0.05

WS_DEFAULT_PORT = 28333
WS_DEFAULT_HOST = LOCALHOST
RPC_SUBSCRIBER_MAX_QUEUE_SIZE = 5000

ETH_GAS_RUNNING_AVERAGE_SIZE = 10000
ADDITIONAL_BLOCKCHAIN_RECONNECT_TIMEOUT_S = 3
CHECK_RELAY_CONNECTIONS_DELAY_S = 5
ETH_MIN_GAS_INTERVAL_S = 5 * 60

ETH_PROTOCOL_VERSION_63 = 63
ONE_DAY_INTERVAL_S = 60 * 60 * 24
ONE_HOUR_INTERVAL_S = 60 * 60


import struct
import time
from typing import Optional

from bxgateway import ont_constants
from bxgateway.messages.ont.ont_message import OntMessage
from bxgateway.messages.ont.ont_message_type import OntMessageType


class VersionOntMessage(OntMessage):
    MESSAGE_TYPE = OntMessageType.VERSION

    def __init__(self, magic: Optional[int] = None, version: Optional[int] = None, sync_port: Optional[int] = None,
                 http_info_port: Optional[int] = None, cons_port: Optional[int] = None, cap: Optional[bytes] = None,
                 nonce: Optional[int] = None, start_height: Optional[int] = None, relay: Optional[int] = None,
                 is_consensus: Optional[bool] = None, soft_version: Optional[bytes] = None,
                 services=ont_constants.ONT_NODE_SERVICES, buf: Optional[bytearray] = None):

        if buf is None:
            # pyre-fixme[6]: Expected `Sized` for 1st param but got `Optional[bytes]`.
            buf = bytearray(ont_constants.ONT_VERSION_MSG_LEN + len(soft_version))
            self.buf = buf

            off = ont_constants.ONT_HDR_COMMON_OFF
            struct.pack_into("<IQQ", buf, off, version, services, int(time.time()))
            off += ont_constants.ONT_VERSION_VER_SERV_TIME_LEN
            struct.pack_into("<HHH", buf, off, sync_port, http_info_port, cons_port)
            off += ont_constants.ONT_VERSION_PORTS_LEN
            # pyre-fixme[6]: Expected `Sized` for 1st param but got `Optional[bytes]`.
            struct.pack_into("<32sQQB?%dp" % (len(soft_version) + 1,), buf, off, cap, nonce,
                             start_height, relay, is_consensus, soft_version)
            # pyre-fixme[6]: Expected `Sized` for 1st param but got `Optional[bytes]`.
            off += ont_constants.ONT_VERSION_LOAD_LEN + len(soft_version)

            super().__init__(magic, self.MESSAGE_TYPE, off - ont_constants.ONT_HDR_COMMON_OFF, buf)

        else:
            self.buf = buf
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._memoryview = memoryview(buf)
        self._version = self._services = self._timestamp = self._sync_port = self._http_info_port = None
        self._cons_port = self._cap = self._nonce = self._start_height = self._relay = self._is_consensus = None
        self._soft_version = None



from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage


class RawEthProtocolMessage(EthProtocolMessage):
    def __init__(self, msg_bytes):
        if not msg_bytes:
            raise ValueError("Message bytes required for raw message")

        super(EthProtocolMessage, self).__init__(msg_bytes)

    def serialize(self):
        raise NotImplementedError("Raw message can't be serialized")

    def deserialize(self):
        raise NotImplementedError("Raw message can't be deserialized")




from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType
from bxutils.logging import LogLevel


class ConfirmedBlockMessage(AbstractBloxrouteMessage):
    """
    Confirmation message that a block has been accepted into the
    blockchain node's blockchain. Only sent to/from streaming gateways.
    """

    MESSAGE_TYPE = GatewayMessageType.CONFIRMED_BLOCK
    MSG_SIZE = (
        AbstractBloxrouteMessage.HEADER_LENGTH
        + crypto.SHA256_HASH_LEN
        + constants.CONTROL_FLAGS_LEN
    )
    PAYLOAD_LENGTH = crypto.SHA256_HASH_LEN + constants.CONTROL_FLAGS_LEN
    EMPTY_BLOCK_CONTENT = memoryview(bytes())

    buf: bytearray
    _block_hash: Optional[Sha256Hash] = None
    _block_content: Optional[memoryview] = None

    def __init__(
        self,
        block_hash: Optional[Sha256Hash] = None,
        block_content: memoryview = EMPTY_BLOCK_CONTENT,
        buf: Optional[bytearray] = None,
    ) -> None:
        if buf is None:
            assert block_hash is not None

            buf = bytearray(self.MSG_SIZE + len(block_content))

            off = AbstractBloxrouteMessage.HEADER_LENGTH
            buf[off:off + crypto.SHA256_HASH_LEN] = block_hash.binary
            off += crypto.SHA256_HASH_LEN
            buf[off:off + len(block_content)] = block_content

        self.buf = buf
        super().__init__(self.MESSAGE_TYPE, self.PAYLOAD_LENGTH + len(block_content), buf)

    def block_hash(self) -> Sha256Hash:
        if self._block_hash is None:
            off = AbstractBloxrouteMessage.HEADER_LENGTH
            self._block_hash = Sha256Hash(self._memoryview[off:off + crypto.SHA256_HASH_LEN])

        block_hash = self._block_hash
        assert block_hash is not None
        return block_hash

    def block_content(self) -> memoryview:
        if self._block_content is None:
            off = AbstractBloxrouteMessage.HEADER_LENGTH + crypto.SHA256_HASH_LEN
            self._block_content = self._memoryview[off: len(self._memoryview) - constants.CONTROL_FLAGS_LEN]

        block_content = self._block_content
        assert block_content is not None
        return block_content

    def log_level(self) -> LogLevel:
        return LogLevel.DEBUG

    def __repr__(self) -> str:
        return "ConfirmedBlockMessage<block_hash: {0}, block_content_size: {1}>".format(
            self.block_hash(), len(self.block_content())
        )



from dataclasses import dataclass
from typing import Optional, Dict

from bxgateway.utils.logging.status.connection_state import ConnectionState
from bxgateway.utils.logging.status.gateway_status import GatewayStatus


@dataclass
class Summary:
    gateway_status: Optional[GatewayStatus] = None
    account_info: Optional[str] = None
    relay_connection_state: Optional[ConnectionState] = None
    blockchain_node_connection_states: Optional[Dict[str, ConnectionState]] = None
    remote_blockchain_node_connection_state: Optional[ConnectionState] = None
    ip_address: Optional[str] = None
    continent: Optional[str] = None
    country: Optional[str] = None
    update_required: bool = False
    quota_level: Optional[str] = None


def gateway_status_get_account_info(account_id: Optional[str]) -> str:
    return f"This gateway is associated to account: {account_id}" if account_id \
        else "This gateway is not registered to any account and is limited to the daily free quota"


def gateway_status_get_quota_level(quota_level: Optional[int]) -> str:
    return f"{quota_level}%"


PROTOCOL_VERSION = 23

SPLIT_RELAYS = 22
TX_MSG_WITH_ACCOUNT_ID = 21
EXPOSE_BDN_LOCAL_REGION = 19
BDN_PERFORMANCE_IGNORE_SEEN = 18
BDN_PERFORMANCE_MULTI_NODE = 17
BDN_PERFORMANCE_NEW_BLOCK = 15

RELAY_BLOCK_CAN_SEND_COMPRESSED_BLOCK_TXS_MESSAGE = 13
RELAY_BLOCK_CAN_SEND_TXS_MESSAGE = 12

# PROTOCOL_VERSION 22 (01/26/2021)
# add to TxMessage account id

# PROTOCOL_VERSION 21 (01/10/2021)
# increase timestamp size to double

# PROTOCOL_VERSION 20 (12/21/2020)
# support additional transaction flag - local region

# PROTOCOL_VERSION 19 (11/18/2020)
# add tx_sent_to_node and duplicate_tx_from_node to BDN performance message
#
# PROTOCOL_VERSION 18 (11/16/2020)
# support additional transaction flag
#
# PROTOCOL_VERSION 17 (10/27/2020)
# support multi-node BDN performance message
#
# PROTOCOL_VERSION 16 (10/15/2020)
# add transaction flag to TxMessage
#
# PROTOCOL_VERSION 15 (9/22/2020)
# add new_block_messages and new_block_announcements to BDN performance messages
#
# PROTOCOL VERSION 14 (9/6/2020)
# Pong(Ping Response contains additional timestamp)
#
# PROTOCOL_VERSION 12 (8/6/2020)
# relay block can send txs_message to gateways
#
# PROTOCOL_VERSION 11 (7/17/2020)
# add memory utilization to BDN performance stats message
#
# PROTOCOL_VERSION 10 (6/8/2020)
# change maximum transaction count in BDN performance stats message
#
# PROTOCOL_VERSION 9 (4/17/2020)
# add broadcast_type to broadcast messages
#
# PROTOCOL_VERSION 8 (2/5/2020)
# add timestamp to tx message types
#
# PROTOCOL_VERSION 7
# added TxQuotaFlag to transaction messages
#
#
# Copy the following template for each protocol version update:
# PROTOCOL_VERSION {N} (MM/DD/YYYY)
# brief description of the changes



from dataclasses import dataclass
from typing import Optional, Dict, Any

from bxcommon import constants
from bxcommon.models.node_type import NodeType


@dataclass
class OutboundPeerModel:
    ip: str
    port: int
    node_id: Optional[str]
    is_internal_gateway: bool
    node_type: Optional[NodeType]
    attributes: Dict[Any, Any]
    non_ssl_port: Optional[int]
    assigning_short_ids: bool
    idx: int

    def __init__(
        self,
        ip: str,
        port: int,
        node_id: Optional[str] = None,
        is_internal_gateway: bool = False,
        attributes: Optional[Dict[Any, Any]] = None,
        node_type: Optional[NodeType] = None,
        non_ssl_port: Optional[int] = None,
        assigning_short_ids: bool = False,
        idx: int = 0,
    ):
        if attributes is None:
            attributes = {}

        self.ip = ip
        self.port = port
        self.node_id = node_id
        self.is_internal_gateway = is_internal_gateway
        self.node_type = node_type
        self.attributes = attributes
        self.non_ssl_port = non_ssl_port
        self.assigning_short_ids = assigning_short_ids
        self.idx = idx

    def get_country(self) -> Optional[str]:
        if constants.NODE_COUNTRY_ATTRIBUTE_NAME in self.attributes:
            return self.attributes[constants.NODE_COUNTRY_ATTRIBUTE_NAME]

        return None

    def get_region(self) -> Optional[str]:
        if constants.NODE_REGION_ATTRIBUTE_NAME in self.attributes:
            return self.attributes[constants.NODE_REGION_ATTRIBUTE_NAME]

        return None

    def is_transaction_streamer(self) -> bool:
        return self.attributes.get(
            constants.TRANSACTION_STREAMER_ATTRIBUTE_NAME, False
        )

    def using_private_ip(self) -> bool:
        return self.attributes.get(
            constants.PRIVATE_IP_ATTRIBUTE_NAME, False
        )

    def __str__(self):
        return (
            f"({self.node_type}, {self.ip}, {self.port}, {self.node_id}, "
            f"{self.is_internal_gateway}, {self.non_ssl_port}, "
            f"{self.attributes}, assigning: {self.assigning_short_ids}, "
            f"idx: {self.idx})"
        )

    def __repr__(self):
        return "OutboundPeerModel" + self.__str__()

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, OutboundPeerModel)
            and other.node_id == self.node_id
            and other.port == self.port
            and other.ip == self.ip
        )

    def __hash__(self):
        return hash(f"{self.node_id}{self.ip}{self.port}")

    @classmethod
    def from_string(cls, peer_info_str: str) -> "OutboundPeerModel":
        try:
            peer_info = peer_info_str.split(":")
            ip = peer_info[0]
            port = peer_info[1]
            node_type = NodeType[peer_info[2].upper()] if len(peer_info) > 2 else NodeType.RELAY
            return OutboundPeerModel(ip, int(port), node_type=node_type)
        except Exception as _e:
            raise ValueError(f"{peer_info_str} is not a valid peer. Specify peer as ip:port:type string.")



from bxcommon.models.quota_type_model import QuotaType
from bxcommon.models.serializable_flag import SerializableFlag


class TransactionFlag(SerializableFlag):
    NO_FLAGS = 0
    STATUS_MONITORING = 1
    PAID_TX = 2
    STATUS_TRACK = STATUS_MONITORING | PAID_TX
    NONCE_MONITORING = 4
    RE_PROPAGATE = 8
    NONCE_TRACK = NONCE_MONITORING | STATUS_TRACK

    # for professional/enterprise accounts only
    CEN_ENABLED = 16

    LOCAL_REGION = 32

    PRIVATE_TX = 64

    # adding more flags to avoid creating new converters
    TBD_2 = 128
    TBD_3 = 256
    TBD_4 = 512
    TBD_5 = 1024

    def get_quota_type(self) -> QuotaType:
        if TransactionFlag.PAID_TX in self:
            return QuotaType.PAID_DAILY_QUOTA
        else:
            return QuotaType.FREE_DAILY_QUOTA



from typing import Iterator, Optional, TypeVar, Generic
from dataclasses import dataclass

from bxcommon.feed.feed_source import FeedSource
from bxcommon.utils.object_hash import Sha256Hash

T = TypeVar("T")


@dataclass
class AbstractRawBlock(Generic[T]):
    """
    AbstractBlock Notification Object

    """
    block_number: int
    block_hash: Sha256Hash
    source: FeedSource
    delayed_block: Iterator[T]
    _block: Optional[T] = None

    @property
    def block(self) -> Optional[T]:
        if self._block is None:
            try:
                self._block = next(self.delayed_block)
            except StopIteration:
                pass
        return self._block



from typing import Tuple, Union, List, Any

from mock import MagicMock

from bxcommon.test_utils import helpers
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway.connections.ont.ont_gateway_node import OntGatewayNode
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import TransactionsEthProtocolMessage
from bxgateway.messages.ont.tx_ont_message import TxOntMessage
from bxgateway.services.gateway_transaction_service import GatewayTransactionService
from bxgateway.testing.abstract_gateway_transaction_service_test import TestAbstractGatewayTransactionService


class TestAbstractOntGatewayTransactionService(TestAbstractGatewayTransactionService):

    def _get_transaction_service(self) -> GatewayTransactionService:
        return GatewayTransactionService(self.node, 0)

    def _get_gateway_node(self) -> AbstractGatewayNode:
        mockSslService = MagicMock()
        return OntGatewayNode(self.opts, mockSslService)

    def _get_node_tx_message(
        self
    ) -> Tuple[
        Union[TxOntMessage, TransactionsEthProtocolMessage], List[Tuple[Any, Any]]
    ]:
        magic = 123456
        version = 1
        tx_contents = helpers.generate_bytearray(200)
        msg = TxOntMessage(magic, version, tx_contents)
        tx_contents = msg.payload()
        return msg, [(msg.tx_hash(), tx_contents)]



from ssl import SSLContext

from cryptography.x509 import Certificate
from mock import MagicMock

from bxutils.services.node_ssl_service import NodeSSLService
from bxutils.ssl.ssl_certificate_type import SSLCertificateType


class MockNodeSSLService(NodeSSLService):

    def blocking_load(self) -> None:
        pass

    def create_ssl_context(self, cert_type: SSLCertificateType) -> SSLContext:
        pass

    def get_certificate(self, cert_type: SSLCertificateType) -> Certificate:
        return MagicMock()



import ipaddress
import blxr_rlp as rlp
class Endpoint(object):
    def serialize(self, obj):
        if not isinstance(obj, tuple):
            raise ValueError("Tuple of values (ip, udp port, tcp port) is expected.")

        ip, udp_port, tcp_port = obj

        ip_address = ipaddress.ip_address(ip)

        return list((ip_address.packed,
                     rlp.sedes.big_endian_int.serialize(udp_port),
                     rlp.sedes.big_endian_int.serialize(tcp_port)))

    def deserialize(self, serialized_data):
        if not isinstance(serialized_data, list) or len(serialized_data) != 3:
            raise ValueError("List with 3 items is expected")

        address = ipaddress.ip_address(serialized_data[0])

        return address.compressed, \
               rlp.sedes.big_endian_int.deserialize(serialized_data[1]), \
               rlp.sedes.big_endian_int.deserialize(serialized_data[2])



import typing
from abc import abstractmethod
from typing import TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.object_hash import Sha256Hash

from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxgateway.messages.ont.get_data_ont_message import GetDataOntMessage
from bxgateway.messages.ont.inventory_ont_message import InventoryOntType
from bxgateway.services.abstract_block_cleanup_service import AbstractBlockCleanupService
from bxgateway.services.ont.ont_block_queuing_service import OntBlockQueuingService

if TYPE_CHECKING:
    from bxgateway.connections.ont.ont_gateway_node import OntGatewayNode

from bxutils import logging

logger = logging.get_logger(__name__)


class AbstractOntBlockCleanupService(AbstractBlockCleanupService):
    """
    Service for managing block cleanup.
    """

    def __init__(self, node: "OntGatewayNode", network_num: int) -> None:
        """
        Constructor
        :param node: reference to node object
        :param network_num: network number
        """

        super(AbstractOntBlockCleanupService, self).__init__(node=node, network_num=network_num)

    @abstractmethod
    def clean_block_transactions(
            self,
            block_msg: BlockOntMessage,
            transaction_service: TransactionService
    ) -> None:
        pass

    def clean_block_transactions_from_block_queue(
            self,
            block_hash: Sha256Hash,
            block_queuing_service: OntBlockQueuingService
    ) -> None:
        if block_hash in block_queuing_service:
            block_msg = self.node.block_queuing_service_manager.get_block_data(block_hash)
            self.node.block_cleanup_service.clean_block_transactions(
                transaction_service=self.node.get_tx_service(),
                block_msg=block_msg
            )
        else:
            logger.debug("block cleanup from queuing service failed, block is no longer tracked {}", block_hash)

    def _request_block(self, block_hash: Sha256Hash) -> None:
        block_request_message = GetDataOntMessage(
            magic=self.node.opts.blockchain_net_magic,
            inv_type=InventoryOntType.MSG_BLOCK.value,
            # pyre-fixme[6]: Expected `[OntObjectHash]` for 3rd parameter but got `Sha256Hash`
            block=block_hash
        )
        node_conn = typing.cast("bxgateway.connections.ont.ont_node_connection.OntNodeConnection",
                                self.node.get_any_active_blockchain_connection())
        if node_conn is None:
            logger.debug("Request for block '{}' failed. No connection to node.", repr(block_hash))
            return
        node_conn.enqueue_msg(block_request_message)
        logger.trace("Received block cleanup request: {}", block_hash)



import fnmatch
from typing import TYPE_CHECKING, Callable, Any, List, Optional

from bxcommon import constants
from bxcommon.feed.feed import FeedKey
from bxcommon.feed.feed_manager import FeedManager
from bxcommon.feed.subscriber import Subscriber
from bxcommon.models.bdn_account_model_base import BdnAccountModelBase
from bxcommon.models.bdn_service_model_config_base import BdnFeedServiceModelConfigBase
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcAccountIdError, RpcError
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)
logger_filters = logging.get_logger(LogRecordType.TransactionFiltering, __name__)


class SubscribeRpcRequest(AbstractRpcRequest["AbstractNode"]):
    help = {
        "params":
        '[feed_name, '
        '{'
        '"include": [field_1, field_2], "duplicates": false, "include_from_blockchain": true, '
        '"blockchain_network": "Mainnet", "blockchain_protocol: "Ethereum"}].\n'
        "Available feeds: newTxs, pendingTxs, newBlocks, txReceipts, onBlock\n"
        "Available fields for transaction feeds: tx_hash, tx_contents (default: all)\n"
        "Available fields for block feed: hash, block (default: all)\n"
        "duplicates: false (filter out duplicates from feed, typically low fee "
        "transactions, default), true (include all duplicates)\n"
        "include_from_blockchain: include transactions received from the connected blockchain node (default: true)\n",
        "description": "Subscribe to a named feed for notifications",
    }

    def __init__(
        self,
        request: BxJsonRpcRequest,
        node: "AbstractNode",
        feed_manager: FeedManager,
        subscribe_handler: Callable[[Subscriber, FeedKey, Optional[str]], None],
        feed_network: int = 0,
        account_details: Optional[BdnAccountModelBase] = None
    ) -> None:
        self.feed_name = ""
        self.feed_network = feed_network
        self.feed_key = FeedKey(self.feed_name, self.feed_network)
        self.feed_manager = feed_manager
        self.subscribe_handler = subscribe_handler
        self.options = {}
        self.available_fields = []
        self.all_fields = []
        self.account_details = account_details
        self.service_model: Optional[BdnFeedServiceModelConfigBase] = None

        super().__init__(request, node)

        assert self.feed_name != ""

    def validate_params(self) -> None:
        try:
            if not self.feed_manager.feeds:
                raise RpcAccountIdError(
                    self.request_id,
                    f"Account does not have access to the transaction streaming service.",
                )

            self.validate_params_get_options()
            self.validate_params_feed_details()
            self.validate_params_service_details()
            self.validate_params_include_fields()
            self.validate_params_filters()
            assert self.feed_name != ""
        except RpcError as e:
            logger.debug({"msg": "Failed to validate subscribe request", "params": self.params, **e.to_json()})
            raise e

    def validate_params_get_options(self):
        params = self.params

        if not isinstance(params, list) or len(params) != 2:
            raise RpcInvalidParams(
                self.request_id,
                "Subscribe RPC request params must be a list of length 2.",
            )
        feed_name, options = params
        if self.feed_network == constants.ALL_NETWORK_NUM:
            self.feed_network = self.node.network_num
        self.feed_name = feed_name
        self.feed_key = FeedKey(self.feed_name, self.feed_network)
        logger.debug("Got new subscribe request for {}", self.feed_key)

        self.options = options

    def validate_params_feed_details(self):
        feed = self.feed_manager.get_feed(self.feed_key)
        if feed is None:
            raise RpcInvalidParams(
                self.request_id,
                f"{self.feed_name} is an invalid feed. "
                f"Available feeds: {[key.name for key in self.feed_manager.get_feed_keys(self.feed_network)]}",
            )
        self.available_fields = feed.FIELDS
        self.all_fields = feed.ALL_FIELDS

    def validate_params_include_fields(self):
        if self.service_model and self.service_model.available_fields:
            self.available_fields = [
                field for field in self.available_fields if allowed_field(field, self.service_model.available_fields)
            ]

        invalid_options = RpcInvalidParams(
            self.request_id,
            f"{self.options} Invalid feed include parameter. "
            "Your plan does not support all requested include parameters "
            'Valid format/fields: {"include": '
            f"{self.available_fields}"
            "}.",
        )
        if not isinstance(self.options, dict):
            raise invalid_options

        include = self.options.get("include", self.all_fields)
        if not isinstance(include, list):
            raise invalid_options
        # check for empty list
        if not include:
            include = self.all_fields

        if self.available_fields:
            if any(
                included_field not in self.available_fields for included_field in include
            ):
                raise invalid_options

            # update options["include"] to support if was not specified
            self.options["include"] = include
        else:
            self.options["include"] = self.available_fields

    def validate_params_filters(self):
        if "filters" in self.options and (not self.service_model or not self.service_model.allow_filtering):
            raise RpcAccountIdError(
                self.request_id,
                f"Account does not have filtering enabled for {self.feed_name} service.",
            )

        filters = self.options.get("filters", None)
        if filters:
            logger_filters.debug(filters)
            formatted_filters = self.format_filters(filters)
            logger_filters.debug("Validated filters: {}", formatted_filters)
            self.options["filters"] = formatted_filters

    def validate_params_service_details(self):
        if self.account_details is None:
            return

        service = self.account_details.get_feed_service_config_by_name(self.feed_name)
        if service:
            service_model = service.feed
        else:
            service_model = None
        if not service or not service.is_service_valid():
            raise RpcAccountIdError(
                self.request_id,
                f"Account does not have access to the {self.feed_name} service.",
            )
        self.service_model = service_model

    async def process_request(self) -> JsonRpcResponse:
        params = self.params
        assert isinstance(params, list)

        subscriber = self.feed_manager.subscribe_to_feed(self.feed_key, self.options)
        assert subscriber is not None  # already validated
        account_id = None
        if self.account_details is not None:
            account_details = self.account_details
            assert account_details is not None
            account_id = account_details.account_id
        self.subscribe_handler(subscriber, self.feed_key, account_id)

        return JsonRpcResponse(self.request_id, subscriber.subscription_id)

    def format_filters(self, filters: Any) -> str:
        valid_filters = self.feed_manager.get_valid_feed_filters(self.feed_key)
        invalid_filters = RpcInvalidParams(
            self.request_id,
            f"{filters} is not a valid set of filters. "
            'Valid format/filters: {"include": '
            f"{valid_filters}"
            "}.",
        )
        if not isinstance(filters, str):
            logger.error("Wrong filter type")
            raise invalid_filters
        if not valid_filters:
            raise invalid_filters
        logger_filters.debug("Validating filters")
        try:
            filters, keys = self.feed_manager.validate_feed_filters(self.feed_key, filters)
        except Exception:
            raise invalid_filters
        # for key in filters, if not in valid_filters, raise
        for key in keys:
            if key not in valid_filters:
                raise RpcInvalidParams(
                    self.request_id,
                    f"{key} is not a valid filter. "
                    'Valid format/filters: {"include": '
                    f"{valid_filters}"
                    "}.",
                )
        return filters


def allowed_field(field: str, available_fields: List[str]):
    for available_field in available_fields:
        if fnmatch.fnmatch(field, available_field) or available_field == "all":
            return True
    return False



from typing import Optional, List, Union, Any, Dict, Tuple

import websockets
import ssl

from bxcommon import constants
from bxcommon.rpc.provider.abstract_ws_provider import AbstractWsProvider
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.rpc_errors import RpcError
from bxcommon.rpc.rpc_request_type import RpcRequestType


class WsProvider(AbstractWsProvider):
    """
    Provider that connects to bxgateway's websocket RPC endpoint.

    Usage:

    (with context manager, recommended)
    ```
    ws_uri = "ws://127.0.0.1:28333"
    async with WsProvider(ws_uri) as ws:
        subscription_id = await ws.subscribe("newTxs")
        while True:
            next_notification = await ws.get_next_subscription_notification_by_id(subscription_id)
            print(next_notification)  # or process it generally
    ```

    (without context manager)
    ```
    ws_uri = "ws://127.0.0.1:28333"
    try:
        ws = await WsProvider(ws_uri)
        subscription_id = await ws.subscribe("newTxs")
        while True:
            next_notification = await ws.get_next_subscription_notification_by_id(subscription_id)
            print(next_notification)  # or process it generally
    except:
        await ws.close()
    ```

    (callback interface)
    ```
    ws_uri = "ws://127.0.0.1:28333"
    ws = WsProvider(ws_uri)
    await ws.initialize()

    def process(subscription_message):
        print(subscription_message)

    ws.subscribe_with_callback(process, "newTxs")

    while True:
        await asyncio.sleep(0)  # otherwise program would exit
    ```
    """

    def __init__(
        self,
        uri: str,
        retry_connection: bool = False,
        queue_limit: int = constants.WS_PROVIDER_MAX_QUEUE_SIZE,
        headers: Optional[Dict] = None,
        skip_ssl_cert_verify: bool = False
    ):
        super(WsProvider, self).__init__(uri, retry_connection, queue_limit, headers)
        if uri.startswith("wss"):
            ssl_context = ssl.SSLContext()
            assert ssl_context is not None
            if skip_ssl_cert_verify:
                ssl_context.check_hostname = False
            self.ssl_context = ssl_context
        else:
            self.ssl_context = None

    async def connect_websocket(self) -> websockets.WebSocketClientProtocol:
        return await websockets.connect(self.uri, extra_headers=self.headers, ssl=self.ssl_context)

    async def call_bx(
        self,
        method: RpcRequestType,
        params: Union[List[Any], Dict[Any, Any], None],
        request_id: Optional[str] = None,
    ) -> JsonRpcResponse:
        if request_id is None:
            request_id = str(self.current_request_id)
            self.current_request_id += 1

        return await self.call(
            BxJsonRpcRequest(request_id, method, params)
        )

    async def subscribe(self, channel: str, options: Optional[Dict[str, Any]] = None) -> str:
        if options is None:
            options = {}
        response = await self.call_bx(
            RpcRequestType.SUBSCRIBE, [channel, options]
        )
        subscription_id = response.result
        assert isinstance(subscription_id, str)
        self.subscription_manager.register_subscription(subscription_id)
        return subscription_id

    async def unsubscribe(self, subscription_id: str) -> Tuple[bool, Optional[RpcError]]:
        response = await self.call_bx(RpcRequestType.UNSUBSCRIBE, [subscription_id])
        if response.result is not None:
            self.subscription_manager.unregister_subscription(subscription_id)
            return True, None
        else:
            return False, response.error



from typing import Dict, Any, Union

from bxcommon.utils.blockchain_utils.eth import transaction_validation_utils
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon import log_messages
from bxcommon.messages.eth.serializers.transaction import Transaction
from bxutils import logging

logger = logging.get_logger(__name__)


class EthTransactionFeedEntry:
    tx_hash: str
    tx_contents: Dict[str, Any]
    local_region: bool

    def __init__(
        self,
        tx_hash: Sha256Hash,
        tx_contents: Union[memoryview, Dict[str, Any]],
        local_region: bool,
    ) -> None:
        self.tx_hash = f"0x{str(tx_hash)}"

        try:
            if isinstance(tx_contents, memoryview):
                transaction = transaction_validation_utils.parse_transaction(tx_contents)
                if transaction is None:
                    raise Exception("could not parse transaction from bytes")
                self.tx_contents = transaction.to_json()
            else:
                # normalize json from source
                self.tx_contents = Transaction.from_json(tx_contents).to_json()
        except Exception as e:
            tx_contents_str = tx_contents
            if isinstance(tx_contents, memoryview):
                tx_contents_str = tx_contents.tobytes()
            logger.error(
                log_messages.COULD_NOT_DESERIALIZE_TRANSACTION,
                tx_hash,
                tx_contents_str,
                e
            )
            raise e

        self.local_region = local_region

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, EthTransactionFeedEntry)
            and other.tx_hash == self.tx_hash
            and other.tx_contents == self.tx_contents
            and other.local_region == self.local_region
        )



import asyncio
import time

from typing import Dict, Any, TYPE_CHECKING, Union, List, Optional
from asyncio import QueueFull, Task
from dataclasses import dataclass, asdict

from bxcommon import constants
from bxcommon.feed.feed_source import FeedSource
from bxcommon.models.serializeable_enum import SerializeableEnum
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcError
from bxcommon.rpc import rpc_constants
from bxcommon.feed.feed import Feed
from bxcommon.feed.subscriber import Subscriber

from bxcommon import log_messages

from bxgateway.utils.stats.eth_on_block_feed_stats_service import (
    eth_on_block_feed_stats_service,
)

from bxutils import logging

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)

SUPPORTED_PAYLOAD_OPTIONS = {"data", "from", "to", "gasPrice", "gas", "address", "pos"}
RETRIES_MAX_ATTEMPTS = 2
RETRIES_SLEEP_INTERVAL = 0.01

TAG_TYPE = Union[str, int]


class EventType(SerializeableEnum):
    TASK_COMPLETED_EVENT = "TaskCompletedEvent"
    TASK_DISABLED_EVENT = "TaskDisabledEvent"


class EthCommandMethod(SerializeableEnum):
    ETH_CALL = "eth_call"  # https://eth.wiki/json-rpc/API#eth_call [{payload}, TAG]
    ETH_GET_BALANCE = "eth_getBalance"  # https://eth.wiki/json-rpc/API#eth_getbalance [address, TAG]
    ETH_GET_TRANSACTION_COUNT = "eth_getTransactionCount"  # https://eth.wiki/json-rpc/API#eth_gettransactioncount
    ETH_GET_STORAGE_AT = "eth_getStorageAt"  # https://eth.wiki/json-rpc/API#eth_getstorageat [address, pos, TAG]
    ETH_GET_CODE = "eth_getCode"  # https://eth.wiki/json-rpc/API#eth_getcode [address, TAG]
    ETH_GET_BLOCK_NUMBER = "eth_blockNumber"  # https://eth.wiki/json-rpc/API#eth_blockNumber []

    @classmethod
    def from_string(cls, value) -> "EthCommandMethod":
        return EthCommandMethod(value)


DEFAULT_METHOD = EthCommandMethod.ETH_CALL


class OnBlockFeedEntry:
    name: str
    response: Dict
    block_height: int
    tag: TAG_TYPE

    def __init__(
        self, name: str, response: Dict, block_height: int, tag: TAG_TYPE
    ) -> None:
        self.name = name
        self.response = response
        self.block_height = block_height
        self.tag = tag


@dataclass()
class EthCallOption:
    command_method: EthCommandMethod
    block_offset: int
    call_name: str
    call_payload: Dict[str, Any]
    active: bool
    
    def validate_item_in_payload(self, item) -> None:
        if item not in self.call_payload:
            raise RpcInvalidParams(
                f"Expected {item} element in request payload for {self.command_method}"
            )

    def __post_init__(self) -> None:
        if self.command_method in {EthCommandMethod.ETH_CALL}:
            self.validate_item_in_payload("data")
        if self.command_method in {
            EthCommandMethod.ETH_GET_BALANCE,
            EthCommandMethod.ETH_GET_CODE,
            EthCommandMethod.ETH_GET_TRANSACTION_COUNT,
            EthCommandMethod.ETH_GET_STORAGE_AT,
        }:
            self.validate_item_in_payload("address")
        if self.command_method in {
            EthCommandMethod.ETH_GET_STORAGE_AT,
        }:
            self.validate_item_in_payload("pos")


class EventNotification:
    block_height: int

    def __init__(self, block_height: int,) -> None:
        self.block_height = block_height


def process_call_params_tag(call_tag: Optional[TAG_TYPE]) -> int:
    if call_tag is None or (isinstance(call_tag, str) and call_tag == "latest"):
        block_offset = 0
    elif isinstance(call_tag, int) and call_tag <= 0:
        block_offset = call_tag
    else:
        raise RpcInvalidParams(
            f"Invalid Value for tag provided {call_tag} use latest, 0 or a negative number"
        )
    return block_offset


def process_call_params_method(method: Optional[str]) -> EthCommandMethod:
    if method is None:
        eth_command_method = DEFAULT_METHOD
    else:
        try:
            eth_command_method = EthCommandMethod.from_string(method)
        except ValueError:
            raise RpcInvalidParams(
                f"Invalid Value for method provided {method} use {[str(item) for item in EthCommandMethod]}"
            )
    return eth_command_method


def process_call_params_payload(call: Dict[str, Any]) -> Dict[str, Any]:
    call_payload = {k: v for (k, v) in call.items() if k in SUPPORTED_PAYLOAD_OPTIONS}
    return call_payload


def process_call_params(call_params: List[Dict[str, Any]]) -> Dict[str, Any]:
    if call_params is None or not isinstance(call_params, list):
        raise RpcInvalidParams("call_params must be a list")
    calls = {}
    for counter, call in enumerate(call_params):

        call_name = call.get("name", str(counter))
        if call_name in calls:
            raise RpcInvalidParams("unique name must be provided for each call")

        block_offset = process_call_params_tag(call.get("tag"))
        eth_command_method = process_call_params_method(call.get("method"))
        call_payload = process_call_params_payload(call)

        calls[call_name] = EthCallOption(
            eth_command_method,
            block_offset,
            call_name,
            call_payload,
            active=call.get("active", True),
        )
    return calls


class EthOnBlockFeed(Feed[OnBlockFeedEntry, EventNotification]):
    NAME = rpc_constants.ETH_ON_BLOCK_FEED_NAME
    FIELDS = ["name", "response", "block_height", "tag"]
    ALL_FIELDS = FIELDS
    VALID_SOURCES = {FeedSource.BLOCKCHAIN_RPC, FeedSource.BLOCKCHAIN_SOCKET}
    last_block_height: int

    def __init__(self, node: "EthGatewayNode", network_num: int = constants.ALL_NETWORK_NUM,) -> None:
        self.node = node
        self.bad_subscribers = set()
        self.last_block_height = 0
        super().__init__(self.NAME, network_num)

    def subscribe(self, options: Dict[str, Any]) -> Subscriber[OnBlockFeedEntry]:
        call_params = options.get("call_params", [])
        calls = process_call_params(call_params)
        subscription_id = options.get("subscription_id")
        if subscription_id:
            subscriber = self.subscribers.get(subscription_id)
            if subscriber is None:
                raise RpcInvalidParams(
                    f"Subscriber id {subscription_id} for feed {self.NAME} not found"
                )

            subscriber.options["calls"].update(calls)
            return subscriber
        else:
            options["calls"] = calls
            return super().subscribe(options)

    def publish(self, raw_message: EventNotification) -> None:
        block_height = raw_message.block_height
        if block_height and block_height <= self.last_block_height:
            # ignore older blocks by height
            logger.debug(
                "Ignore EthOnBlockFeed block notification for old block height: {} last: {}",
                block_height, self.last_block_height
            )
            return
        logger.debug("Processing EthOnBlockFeed notification for block height: {}", block_height)
        self.last_block_height = block_height
        event_init_time = time.time()

        # remove bad subscribers
        while self.bad_subscribers:
            self.unsubscribe(self.bad_subscribers.pop())

        # process subscriptions
        for subscriber in self.subscribers.values():
            subscriber_tasks = []
            for call in subscriber.options["calls"].values():
                if call.active:
                    task = asyncio.create_task(
                        self._publish(subscriber, call, block_height)
                    )
                    subscriber_tasks.append(task)
            asyncio.create_task(
                self.wait_for_all_subscriber_tasks(
                    subscriber, subscriber_tasks, block_height, event_init_time
                )
            )

    async def wait_for_all_subscriber_tasks(
        self,
        subscriber: Subscriber[OnBlockFeedEntry],
        subscriber_tasks: List[Task],
        block_height: int,
        event_init_time: float,
    ) -> None:
        logger.trace(
            "Execution Completed for block height {} duration {} s",
            block_height,
            time.time() - event_init_time,
        )
        eth_on_block_feed_stats_service.log_subscriber_tasks(
            len(subscriber_tasks), time.time() - event_init_time
        )
        if subscriber_tasks:
            await asyncio.wait(subscriber_tasks, return_when=asyncio.ALL_COMPLETED)
        subscriber.queue(
            self.serialize_response(
                str(EventType.TASK_COMPLETED_EVENT), {}, block_height, block_height
            )
        )

    async def _publish(
        self,
        subscriber: Subscriber[OnBlockFeedEntry],
        call: EthCallOption,
        block_height: int,
        retry_count: int = 0
    ) -> None:
        tag = block_height + call.block_offset
        try:
            response = await self.execute_eth_call(call, tag)
        except RpcError as e:
            response = e.to_json()
            if e.message == "header not found" and retry_count <= RETRIES_MAX_ATTEMPTS:
                await asyncio.sleep(RETRIES_SLEEP_INTERVAL)
                await self._publish(subscriber, call, block_height, retry_count + 1)
                retry = True
            else:
                call.active = False
                subscriber.queue(
                    self.serialize_response(
                        str(EventType.TASK_DISABLED_EVENT),
                        asdict(call),
                        block_height,
                        block_height,
                    )
                )
                retry = False
            logger.info(
                "{}, Error response from node {}, call details {} retry: {}",
                self,
                response,
                call,
                retry,
            )

        serialized_message = self.serialize_response(
            call.call_name, response, block_height, tag
        )
        try:
            subscriber.queue(serialized_message)
        except QueueFull:
            logger.info(
                log_messages.BAD_FEED_SUBSCRIBER, subscriber.subscription_id, self.name
            )
            self.bad_subscribers.add(subscriber.subscription_id)

    def serialize(self, raw_message: EventNotification) -> OnBlockFeedEntry:
        raise NotImplementedError

    def serialize_response(
        self, name: str, response: Dict, block_height: int, tag: TAG_TYPE
    ) -> OnBlockFeedEntry:
        sanitized_response = {
            k: v for (k, v) in response.items() if k not in rpc_constants.ETH_RPC_INTERNAL_RESPONSE_ITEMS
        }
        return OnBlockFeedEntry(name, sanitized_response, block_height, tag)

    async def execute_eth_call(
        self, call_option: EthCallOption, tag: TAG_TYPE
    ) -> Dict:
        if isinstance(tag, int):
            tag = hex(tag)
        payload = call_option.call_payload
        command = call_option.command_method
        if command == EthCommandMethod.ETH_CALL:
            request_payload = [payload, tag]
        elif command in {
            EthCommandMethod.ETH_GET_BALANCE,
            EthCommandMethod.ETH_GET_CODE,
            EthCommandMethod.ETH_GET_TRANSACTION_COUNT,
        }:
            request_payload = [payload["address"], tag]
        elif command in {
            EthCommandMethod.ETH_GET_STORAGE_AT,
        }:
            request_payload = [
                payload["address"],
                payload["pos"],
                tag,
            ]
        elif command in {EthCommandMethod.ETH_GET_BLOCK_NUMBER}:
            request_payload = []
        else:
            raise ValueError(f"Invalid EthCommand Option: {command}")
        response = await self.node.eth_ws_proxy_publisher.call_rpc(
            str(command), request_payload,
        )
        return response.to_json()



from abc import ABCMeta, abstractmethod
from typing import Union, Tuple, List

from bxcommon import constants
from bxcommon.test_utils import helpers
from bxcommon.test_utils.abstract_transaction_service_test_case import AbstractTransactionServiceTestCase
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import TransactionsEthProtocolMessage
from bxgateway.messages.ont.tx_ont_message import TxOntMessage
from bxgateway.services.gateway_transaction_service import GatewayTransactionService
from bxgateway.testing import gateway_helpers


class TestAbstractGatewayTransactionService(AbstractTransactionServiceTestCase, metaclass=ABCMeta):

    def setUp(self) -> None:
        pub_key = "a04f30a45aae413d0ca0f219b4dcb7049857bc3f91a6351288cce603a2c9646294a02b987bf6586b370b2c22d74662355677007a14238bb037aedf41c2d08866"

        self.opts = gateway_helpers.get_gateway_opts(
            8000,
            include_default_eth_args=True,
            pub_key=pub_key,
            blockchain_network_num=5,
            account_id="12345"
        )

        if self.opts.use_extensions:
            helpers.set_extensions_parallelism()
        self.node = self._get_gateway_node()
        self.transaction_service = self._get_transaction_service()

    def _test_process_transactions_message_from_node(self):
        min_tx_network_fee = 0
        self.node.account_id = "12345"
        if self.node.network_num in self.node.opts.blockchain_networks:
            min_tx_network_fee = self.node.opts.blockchain_networks[self.node.network_num].min_tx_network_fee

        tx_message, test_txs_info = self._get_node_tx_message()
        self.opts.transaction_validation = False
        result = self.transaction_service.process_transactions_message_from_node(
            tx_message, min_tx_network_fee, True
        )

        self.assertIsNotNone(result)
        self.assertEqual(len(test_txs_info), len(result))

        total_size = 0

        for (test_tx_hash, test_tx_contents), tx_result in zip(test_txs_info, result):
            transaction_key = self.transaction_service.get_transaction_key(test_tx_hash)
            self.assertFalse(tx_result.seen)
            self.assertEqual(convert.bytes_to_hex(test_tx_hash.binary),
                             convert.bytes_to_hex(tx_result.transaction_hash.binary))
            tx_contents = tx_result.transaction_contents
            self.assertEqual(convert.bytes_to_hex(test_tx_contents), convert.bytes_to_hex(tx_contents))

            self.assertTrue(self.transaction_service.has_transaction_contents(test_tx_hash))
            self.assertEqual(convert.bytes_to_hex(test_tx_contents),
                             convert.bytes_to_hex(self.transaction_service.get_transaction_by_key(transaction_key)))

            total_size += len(test_tx_contents)

        self.assertEqual(0, len(self.transaction_service._tx_assignment_expire_queue.queue))
        self.assertEqual(len(result), len(self.transaction_service.tx_hashes_without_short_id))
        self.assertEqual(0, len(self.transaction_service.tx_hashes_without_content))
        self.assertEqual(total_size, self.transaction_service._total_tx_contents_size)


    @abstractmethod
    def _get_transaction_service(self) -> GatewayTransactionService:
        pass

    @abstractmethod
    def _get_gateway_node(self) -> AbstractGatewayNode:
        pass

    @abstractmethod
    def _get_node_tx_message(
        self
    ) -> Tuple[Union[TxOntMessage, TransactionsEthProtocolMessage], List[Tuple[Sha256Hash, Union[bytearray, memoryview]]]]:
        pass



class OntMessageType:
    VERSION = b"version"
    VERACK = b"verack"
    GET_ADDRESS = b"getaddr"
    ADDRESS = b"addr"
    PING = b"ping"
    PONG = b"pong"
    GET_HEADERS = b"getheaders"
    HEADERS = b"headers"
    INVENTORY = b"inv"
    GET_DATA = b"getdata"
    BLOCK = b"block"
    TRANSACTIONS = b"tx"
    CONSENSUS = b"consensus"
    GET_BLOCKS = b"getblocks"
    NOT_FOUND = b"notfound"
    DISCONNECT = b"disconnect"



import time
from abc import abstractmethod, ABCMeta
from typing import Optional, TYPE_CHECKING, List, Generic, Callable, cast

from bxcommon import constants
from bxcommon.utils.alarm_queue import AlarmId
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway import gateway_constants
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.services.abstract_block_queuing_service import (
    AbstractBlockQueuingService,
    TBlockMessage,
    THeaderMessage,
)
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

logger = logging.get_logger(__name__)


class PushBlockQueuingService(
    AbstractBlockQueuingService[TBlockMessage, THeaderMessage],
    Generic[TBlockMessage, THeaderMessage],
    metaclass=ABCMeta,
):
    """
    Service class with following responsibilities:

    1. Make sure that gateway does not send blocks to blockchain node too fast
       if the previous block has not yet been accepted. This condition is
       ignored if not blocks have been sent yet, or the maximum validation
       timeout has already passed.

    2. Make sure that no new blocks are sent to blockchain node while gateway
       is waiting for transactions corresponding to shorts ids in the previous
       block if it was not able to find them in local tx service cache

    This implementation actively pushes new blocks to the blockchain node
    (as opposed to a "pull-based" approach that sends announcements and lets
    blockchain nodes request them).
    """

    def __init__(
        self,
        node: "AbstractGatewayNode",
        connection: AbstractGatewayBlockchainConnection
    ):
        super(PushBlockQueuingService, self).__init__(node, connection)

        self._last_block_sent_time: float = 0.0
        self._last_alarm_id: Optional[AlarmId] = None

    @abstractmethod
    def get_previous_block_hash_from_message(
        self, block_message: TBlockMessage
    ) -> Sha256Hash:
        pass

    @abstractmethod
    def on_block_sent(
        self, block_hash: Sha256Hash, block_message: TBlockMessage
    ):
        pass

    def push(
        self,
        block_hash: Sha256Hash,
        block_msg: Optional[TBlockMessage] = None,
        waiting_for_recovery: bool = False,
    ) -> None:
        super().push(block_hash, block_msg, waiting_for_recovery)

        if (
            not waiting_for_recovery and
            self.node.has_active_blockchain_peer()
        ):
            assert block_msg is not None
            if self._can_send_block_message(block_hash, block_msg):
                self.send_block_to_node(block_hash, block_msg)
                self.remove_from_queue(block_hash)
                return

        logger.debug(
            "Queued up block {} for sending to the blockchain node {}. "
            "Block is behind {} others.",
            block_hash,
            self.connection.peer_desc,
            len(self._block_queue) - 1,
        )

        # if it is the first item in the queue then schedule alarm
        if len(self._block_queue) == 1:
            self._schedule_alarm_for_next_item()

    def update_recovered_block(
        self, block_hash: Sha256Hash, block_msg: TBlockMessage
    ) -> None:
        if block_hash in self._blocks:
            self._blocks_waiting_for_recovery[block_hash] = False

            # if this is the first item in the queue then cancel alarm for
            # block recovery timeout and send block
            if len(self._block_queue) > 0 and self._block_queue[0][0] == block_hash:
                last_alarm_id = self._last_alarm_id
                assert last_alarm_id is not None
                self.node.alarm_queue.unregister_alarm(last_alarm_id)
                self._schedule_alarm_for_next_item()

    def mark_blocks_seen_by_blockchain_node(
        self, block_hashes: List[Sha256Hash]
    ) -> None:
        """
        Marks blocks seen and retries the top block(s).
        """
        for block_hash in block_hashes:
            self.mark_block_seen_by_blockchain_node(block_hash, None)
        self._retry_send()

    def mark_block_seen_by_blockchain_node(
        self,
        block_hash: Sha256Hash,
        block_message: Optional[TBlockMessage],
        block_number: Optional[int] = None
    ) -> None:
        super().mark_block_seen_by_blockchain_node(block_hash, block_message)
        self._blocks_seen_by_blockchain_node.add(block_hash)
        self.remove_from_queue(block_hash)

    def send_block_to_node(
        self, block_hash: Sha256Hash, block_msg: Optional[TBlockMessage] = None
    ) -> None:
        if block_msg is None:
            block_msg = self.node.block_storage[block_hash]

        super(PushBlockQueuingService, self).send_block_to_node(
            block_hash, block_msg
        )
        self._last_block_sent_time = time.time()
        # pyre-fixme[6]: Expected `TBlockMessage` for 2nd param but got
        #  `Optional[Variable[TBlockMessage (bound to
        #  bxcommon.messages.abstract_message.AbstractMessage)]]`.
        self.on_block_sent(block_hash, block_msg)

    def remove_from_queue(self, block_hash: Sha256Hash) -> int:
        index = super().remove_from_queue(block_hash)

        if index == 0 and self._last_alarm_id is not None:
            # pyre-fixme[6]: Expected `AlarmId` for 1st param but got
            #  `Optional[AlarmId]`.
            self.node.alarm_queue.unregister_alarm(self._last_alarm_id)
            self._schedule_alarm_for_next_item()

        return index

    def _can_send_block_message(
        self, block_hash: Sha256Hash, block_message: TBlockMessage
    ) -> bool:
        """
        Determines if a block can be sent.

        The first block on the queue on node startup (whether it gets there from
        being the first pushed onto the queue or because the actually first
        block failed block recovery) will always get sent.

        Otherwise, wait for previous block hash to be seen to send the block.
        """
        if (
            (not self._block_queue or self._block_queue[0][0] == block_hash)
            and len(self._blocks_seen_by_blockchain_node) == 0
            and self._last_block_sent_time == 0.0
        ):
            return True
        else:
            previous_block_hash = self.get_previous_block_hash_from_message(
                block_message
            )
            return previous_block_hash in self._blocks_seen_by_blockchain_node

    def _retry_send(self) -> None:
        """
        Checks if top block can be sent to blockchain node, sending if so.
        This method should be called after adding blocks to the
        set of previous blocks.
        """
        last_alarm_id = self._last_alarm_id
        if last_alarm_id:
            self.node.alarm_queue.unregister_alarm(last_alarm_id)
        self._schedule_alarm_for_next_item()

    def _schedule_alarm_for_next_item(self) -> None:
        self._last_alarm_id = None

        if len(self._block_queue) == 0:
            return

        block_hash, timestamp = self._block_queue[0]
        waiting_recovery = self._blocks_waiting_for_recovery[block_hash]

        if waiting_recovery:
            timeout = gateway_constants.BLOCK_RECOVERY_MAX_QUEUE_TIME - (
                time.time() - timestamp
            )
            self._run_or_schedule_alarm(
                timeout, self._top_block_recovery_timeout
            )
        else:
            if not self._is_elapsed_time_shorter_than_ttl(time.time() - timestamp, block_hash):
                self._remove_old_blocks_from_queue()
                self._schedule_alarm_for_next_item()
                return
            block_message = cast(TBlockMessage, self.node.block_storage[block_hash])
            assert block_message is not None
            if self._can_send_block_message(block_hash, block_message):
                timeout = 0
            else:
                timeout = self.node.opts.max_block_interval_s - (
                    time.time() - self._last_block_sent_time
                )
            self._run_or_schedule_alarm(timeout, self._send_top_block_to_node)

    def _run_or_schedule_alarm(self, timeout: float, func: Callable) -> None:
        if timeout > 0:
            self._last_alarm_id = self.node.alarm_queue.register_alarm(
                timeout, func
            )
        elif not self.node.has_active_blockchain_peer():
            self.node.alarm_queue.register_alarm(
                gateway_constants.NODE_READINESS_FOR_BLOCKS_CHECK_INTERVAL_S,
                func,
            )
        else:
            func()

    def _send_top_block_to_node(self) -> int:
        if len(self._block_queue) == 0:
            return constants.CANCEL_ALARMS

        if not self.node.has_active_blockchain_peer():
            self._schedule_alarm_for_next_item()
            return constants.CANCEL_ALARMS

        block_hash, timestamp = self._block_queue[0]
        waiting_recovery = self._blocks_waiting_for_recovery[block_hash]

        if waiting_recovery:
            logger.debug(
                "Unable to send block to node {}, requires recovery. "
                "Block hash {}.",
                self.connection.peer_desc,
                block_hash,
            )
            self._schedule_alarm_for_next_item()
            return constants.CANCEL_ALARMS

        block_msg = cast(TBlockMessage, self.node.block_storage[block_hash])
        index = self.remove_from_queue(block_hash)
        assert index == 0

        if self._is_elapsed_time_shorter_than_ttl(time.time() - timestamp, block_hash):
            self.send_block_to_node(block_hash, block_msg)
        else:
            self._remove_old_blocks_from_queue()

        self._schedule_alarm_for_next_item()
        return constants.CANCEL_ALARMS

    def _top_block_recovery_timeout(self) -> int:
        if len(self._block_queue) == 0:
            return constants.CANCEL_ALARMS

        current_time = time.time()

        block_hash, timestamp = self._block_queue[0]
        waiting_block_recovery = self._blocks_waiting_for_recovery[block_hash]

        if not waiting_block_recovery:
            logger.debug(
                "Unable to cancel recovery for block {} in queuing service for {}. "
                "Block is not in recovery.",
                block_hash,
                self.connection.peer_desc
            )
            self._schedule_alarm_for_next_item()
            return constants.CANCEL_ALARMS

        if (
            current_time - timestamp
            < gateway_constants.BLOCK_RECOVERY_MAX_QUEUE_TIME
        ):
            logger.debug(
                "Unable to cancel recovery for block {} in queuing service for {}. "
                "Block recovery did not timeout.",
                block_hash,
                self.connection.peer_desc
            )
            self._schedule_alarm_for_next_item()
            return constants.CANCEL_ALARMS

        index = self.remove(block_hash)
        assert index == 0

        self._schedule_alarm_for_next_item()

        return constants.CANCEL_ALARMS

    def _is_elapsed_time_shorter_than_ttl(self, elapsed_time: float, block_hash: Sha256Hash) -> bool:
        if elapsed_time < self.node.opts.blockchain_message_ttl:
            return True

        logger.debug("Skipping block {}, since {:.2f}s have passed since receiving the block.",
                     block_hash, elapsed_time)
        return False

    def _remove_old_blocks_from_queue(self) -> None:
        """
        Removes old blocks from block queue even if block_hash is not in self._blocks, similar to remove_from_queue
        """
        logger.trace("Checking block queue for {} and removing old blocks from queue.", self.connection.peer_desc)
        current_time = time.time()
        block_queue_start_len = len(self._block_queue)
        while len(self._block_queue) > 0:
            block_hash, timestamp = self._block_queue[0]
            if current_time - timestamp < self.node.opts.blockchain_message_ttl:
                return

            last_alarm_id = self._last_alarm_id
            if block_queue_start_len == len(self._block_queue) and last_alarm_id is not None:
                self.node.alarm_queue.unregister_alarm(last_alarm_id)
                self._schedule_alarm_for_next_item()

            del self._block_queue[0]
            del self._blocks_waiting_for_recovery[block_hash]



from typing import Optional

import blxr_rlp as rlp

from bxcommon.utils.blockchain_utils.eth import eth_common_constants


# pyre-fixme[13]: Attribute `data` is never initialized.
# pyre-fixme[13]: Attribute `gas_price` is never initialized.
# pyre-fixme[13]: Attribute `nonce` is never initialized.
# pyre-fixme[13]: Attribute `start_gas` is never initialized.
# pyre-fixme[13]: Attribute `to` is never initialized.
# pyre-fixme[13]: Attribute `value` is never initialized.
class UnsignedTransaction(rlp.Serializable):
    fields = [
        ("nonce", rlp.sedes.big_endian_int),
        ("gas_price", rlp.sedes.big_endian_int),
        ("start_gas", rlp.sedes.big_endian_int),
        ("to", rlp.sedes.Binary.fixed_length(eth_common_constants.ADDRESS_LEN, allow_empty=True)),
        ("value", rlp.sedes.big_endian_int),
        ("data", rlp.sedes.binary),
    ]

    nonce: int
    gas_price: int
    start_gas: int
    to: Optional[bytearray]
    value: int
    data: bytearray



from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType


class DisconnectRelayPeerMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.DISCONNECT_RELAY_PEER

    def __init__(self, buf=None) -> None:
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + constants.CONTROL_FLAGS_LEN)
            self.buf = buf

            super(DisconnectRelayPeerMessage, self).__init__(self.MESSAGE_TYPE, constants.CONTROL_FLAGS_LEN, self.buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._command = self._payload_len = None
            self._payload = None



import struct
from typing import Optional, Union

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_broadcast_message import AbstractBroadcastMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.models.quota_type_model import QuotaType
from bxcommon.utils.object_hash import Sha256Hash


class TxMessageV7(AbstractBroadcastMessage):
    PAYLOAD_LENGTH = AbstractBroadcastMessage.PAYLOAD_LENGTH + constants.SID_LEN + constants.QUOTA_FLAG_LEN
    MESSAGE_TYPE = BloxrouteMessageType.TRANSACTION
    EMPTY_TX_VAL = memoryview(bytes())

    def __init__(self, message_hash: Optional[Sha256Hash] = None, network_num: Optional[int] = None,
                 source_id: str = "", short_id: int = constants.NULL_TX_SID,
                 tx_val: Union[bytearray, bytes, memoryview, None] = None,
                 quota_type: Optional[QuotaType] = None,
                 buf: Optional[bytearray] = None):
        self._short_id = None
        self._tx_val: Optional[memoryview] = None
        self._tx_quota_type = None

        # override payload length for variable length message
        if tx_val is not None:
            # pylint: disable=invalid-name
            self.PAYLOAD_LENGTH = AbstractBroadcastMessage.PAYLOAD_LENGTH + constants.SID_LEN + \
                                  constants.QUOTA_FLAG_LEN + len(tx_val)
        super().__init__(message_hash, network_num, source_id, buf)

        if buf is None:
            # minus control flag
            off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - constants.CONTROL_FLAGS_LEN

            struct.pack_into("<L", self.buf, off, short_id)
            off += constants.SID_LEN

            if quota_type is None:
                quota_type = QuotaType.FREE_DAILY_QUOTA
            struct.pack_into("<B", self.buf, off, quota_type.value)
            off += constants.QUOTA_FLAG_LEN

            if tx_val is not None:
                self.buf[off:off + len(tx_val)] = tx_val

    def tx_hash(self) -> Sha256Hash:
        return self.message_hash()

    def short_id(self) -> int:
        if self._short_id is None:
            off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - constants.CONTROL_FLAGS_LEN
            self._short_id, = struct.unpack_from("<L", self.buf, off)
        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._short_id

    def quota_type(self) -> QuotaType:
        if self._tx_quota_type is None:
            off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH + constants.SID_LEN - \
                  constants.CONTROL_FLAGS_LEN

            tx_quota_type_flag, = struct.unpack_from("<B", self.buf, off)
            self._tx_quota_type = QuotaType(tx_quota_type_flag)
        assert self._tx_quota_type is not None
        # pyre-fixme[7]: Expected `QuotaType` but got `None`.
        return self._tx_quota_type

    def tx_val(self) -> memoryview:
        if self._tx_val is None:
            if self.payload_len() == 0:
                self._tx_val = self.EMPTY_TX_VAL
            else:
                off = self.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH + constants.SID_LEN + \
                      constants.QUOTA_FLAG_LEN - constants.CONTROL_FLAGS_LEN
                self._tx_val = self._memoryview[
                               off:self.HEADER_LENGTH + self.payload_len() - constants.CONTROL_FLAGS_LEN]

        assert self._tx_val is not None
        # pyre-fixme[7]: Expected `memoryview` but got `Optional[memoryview]`.
        return self._tx_val

    def __repr__(self) -> str:
        return ("TxMessage<tx_hash: {}, short_id: {}, network_num: {}, compact: {}, source_id: {}, quota_type: {}>"
                .format(self.tx_hash(), self.short_id(), self.network_num(), self.tx_val() == self.EMPTY_TX_VAL,
                        self.source_id_as_str(), self.quota_type()))



import json
import os
import traceback
from datetime import date, time, datetime
from enum import Enum
from inspect import istraceback
from typing import Union, Any, Iterable, Collection, Optional, Dict

from bxutils import logging

logger = logging.get_logger(__name__)

SPECIAL_ITERABLE_TYPES = (type(dict().values()), type(dict().keys()),)


class Case(Enum):
    SNAKE = 1
    CAMEL = 2


def is_iterable_no_collection(obj):
    return isinstance(obj, SPECIAL_ITERABLE_TYPES) or \
           (isinstance(obj, Iterable) and not isinstance(obj, Collection))


def to_json(obj: Any, remove_nulls: bool = False) -> str:
    if remove_nulls:
        clean_dict = {}
        for key, value in obj.__dict__.items():
            if value:
                clean_dict[key] = value
        return json.dumps(clean_dict, cls=EnhancedJSONEncoder)

    return json.dumps(obj, cls=EnhancedJSONEncoder)


def to_dict(obj: Any) -> Dict[str, Any]:
    return EnhancedJSONEncoder().as_dict(obj)


def load_json_from_file(json_file_path: str) -> Optional[Union[list, dict]]:
    node_json = None

    if os.path.isfile(json_file_path):
        try:
            with open(json_file_path) as json_file:
                node_json = json.load(json_file)
        except ValueError as e:
            logger.debug("Failed to parse json: {}", e)
        except OSError as e:
            logger.debug("Failed trying to check for a json file: {}", e)
    else:
        raise ValueError(f"Could not locate json file: {json_file_path}")

    return node_json


class EnhancedJSONEncoder(json.JSONEncoder):
    # pylint: disable=method-hidden,too-many-return-statements,arguments-differ
    # pyre-fixme[14]: `default` overrides method defined in `JSONEncoder`
    #  inconsistently.
    def default(self, obj: Any) -> Any:
        if is_iterable_no_collection(obj):
            obj = list(obj)
        elif isinstance(obj, (bytearray, memoryview)):
            obj = bytes(obj)
        if isinstance(obj, Enum):
            return str(obj)
        if hasattr(obj, "__dict__"):
            if isinstance(obj.__dict__, dict):
                return obj.__dict__
            else:
                return str(obj)
        if isinstance(obj, (date, datetime, time)):
            return obj.isoformat()
        if isinstance(obj, bytes):
            try:
                return obj.decode("utf-8")
            except UnicodeDecodeError:
                return str(obj)
        if hasattr(obj, "hexdigest"):
            return obj.hexdigest()
        if hasattr(obj, "hex_string"):
            return obj.hex_string()
        if istraceback(obj):
            return "".join(traceback.format_tb(obj)).strip()
        return obj

    def _encode(self, obj):
        obj = self.default(obj)
        if isinstance(obj, dict):
            return {self.default(self._encode(k)): self._encode(v) for k, v in obj.items()}
        elif isinstance(obj, (list, set)):
            return [self._encode(l) for l in obj]
        else:
            return obj

    def encode(self, o) -> str:
        return super(EnhancedJSONEncoder, self).encode(self._encode(o))

    def as_dict(self, obj) -> Dict[str, Any]:
        return self._encode(obj)



from typing import NamedTuple

from bxcommon import constants
from bxcommon.feed.feed import Feed
from bxcommon.models.tx_blockchain_status import TxBlockchainStatus
from bxcommon.utils.object_hash import Sha256Hash


class TransactionStatusFeedEntry:
    tx_hash: str
    status: str
    account_id: str

    def __init__(
        self,
        tx_hash: Sha256Hash,
        status: TxBlockchainStatus,
        account_id: str
    ) -> None:
        self.tx_hash = str(tx_hash)
        self.status = str(status)
        self.account_id = account_id

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, TransactionStatusFeedEntry)
            and other.tx_hash == self.tx_hash
            and other.status == self.status
            and other.account_id == self.account_id
        )


class TransactionStatus(NamedTuple):
    tx_hash: Sha256Hash
    status: TxBlockchainStatus
    account_id: str


class TransactionStatusFeed(Feed[TransactionStatusFeedEntry, TransactionStatus]):
    NAME = "transactionStatus"
    FIELDS = ["tx_hash", "status", "account_id"]
    ALL_FIELDS = FIELDS

    def __init__(self, network_num: int = constants.ALL_NETWORK_NUM) -> None:
        super().__init__(self.NAME, network_num)

    def serialize(self, raw_message: TransactionStatus) -> TransactionStatusFeedEntry:
        return TransactionStatusFeedEntry(raw_message.tx_hash, raw_message.status, raw_message.account_id)



import weakref
from typing import TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.btc.btc_remote_connection_protocol import BtcRemoteConnectionProtocol
from bxgateway.messages.btc.btc_message_factory import btc_message_factory
from bxgateway.messages.btc.ping_btc_message import PingBtcMessage

if TYPE_CHECKING:
    from bxgateway.connections.btc.btc_gateway_node import BtcGatewayNode


class BtcRemoteConnection(AbstractGatewayBlockchainConnection["BtcGatewayNode"]):
    CONNECTION_TYPE = ConnectionType.REMOTE_BLOCKCHAIN_NODE

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "BtcGatewayNode"):
        super().__init__(sock, node)
        self.connection_protocol = weakref.ref(BtcRemoteConnectionProtocol(self))

    def connection_message_factory(self) -> AbstractMessageFactory:
        return btc_message_factory

    def ping_message(self) -> AbstractMessage:
        return PingBtcMessage(self.node.opts.blockchain_net_magic)



from dataclasses import dataclass
from enum import Enum
from functools import total_ordering
from typing import Optional
from datetime import datetime, date

from bxcommon.models.bdn_service_model_config_base import (
    BdnBasicServiceModel,
    BdnQuotaServiceModelConfigBase,
    BdnFeedServiceModelConfigBase, BdnPrivateRelayServiceModelConfigBase,
)
from bxcommon import constants
from bxcommon.rpc import rpc_constants
from bxutils import logging

logger = logging.get_logger(__name__)
OPTIONAL_ACCOUNT_SERVICES = {"tx_free"}


@total_ordering
class Tiers(Enum):
    INTRODUCTORY = "Introductory"
    DEVELOPER = "Developer"
    PROFESSIONAL = "Professional"
    ENTERPRISE = "Enterprise"
    ENTERPRISE_ELITE = "EnterpriseElite"

    def __eq__(self, other):
        if not isinstance(other, Tiers):
            return NotImplemented

        # pylint: disable=comparison-with-callable
        return self.value == other.value

    def __hash__(self):
        return id(self)

    def __lt__(self, other):
        if not isinstance(other, Tiers):
            return NotImplemented

        order = [
            Tiers.INTRODUCTORY,
            Tiers.DEVELOPER,
            Tiers.PROFESSIONAL,
            Tiers.ENTERPRISE,
            Tiers.ENTERPRISE_ELITE
        ]

        return order.index(self) < order.index(other)

    def __gt__(self, other):
        return not self.__lt__(other)

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    @classmethod
    def from_string(cls, value: str) -> Optional["Tiers"]:
        try:
            return Tiers(value)
        except ValueError:
            return None


@dataclass
class AccountInfo:
    account_id: str
    logical_account_name: str
    certificate: str
    # TODO change expire_date to datetime type
    expire_date: str = constants.EPOCH_DATE
    blockchain_protocol: Optional[str] = None
    blockchain_network: Optional[str] = None
    tier_name: Optional[str] = None
    is_miner: Optional[bool] = None


@dataclass
class AccountTemplate:
    # All items except tx_free are required.
    # Service Model is enabled/disabled according to the service expiry date, default disabled
    tx_free: Optional[BdnQuotaServiceModelConfigBase] = None
    tx_paid: Optional[BdnQuotaServiceModelConfigBase] = None
    block_paid: Optional[BdnQuotaServiceModelConfigBase] = None
    cloud_api: Optional[BdnBasicServiceModel] = None
    new_transaction_streaming: Optional[BdnFeedServiceModelConfigBase] = None
    new_block_streaming: Optional[BdnFeedServiceModelConfigBase] = None
    new_pending_transaction_streaming: Optional[BdnFeedServiceModelConfigBase] = None
    transaction_state_feed: Optional[BdnFeedServiceModelConfigBase] = None
    on_block_feed: Optional[BdnFeedServiceModelConfigBase] = None
    transaction_receipts_feed: Optional[BdnFeedServiceModelConfigBase] = None
    private_relays: Optional[BdnPrivateRelayServiceModelConfigBase] = None
    private_transaction: Optional[BdnQuotaServiceModelConfigBase] = None
    private_transaction_fee: Optional[BdnQuotaServiceModelConfigBase] = None
    light_gateway: Optional[BdnBasicServiceModel] = None
    online_gateways: Optional[BdnQuotaServiceModelConfigBase] = None


@dataclass
class BdnAccountModelBase(AccountTemplate, AccountInfo):

    def is_account_valid(self) -> bool:
        today = datetime.utcnow().date()
        try:
            expire_date = date.fromisoformat(self.expire_date)
        except (KeyError, ValueError):
            return False

        return expire_date >= today

    # pylint: disable=too-many-return-statements
    def get_feed_service_config_by_name(
        self, feed_name: str
    ) -> Optional[BdnFeedServiceModelConfigBase]:
        if feed_name in {rpc_constants.NEW_TRANSACTION_FEED_NAME}:
            return self.new_transaction_streaming
        elif feed_name in {rpc_constants.ETH_PENDING_TRANSACTION_FEED_NAME}:
            return self.new_pending_transaction_streaming
        elif feed_name in {rpc_constants.NEW_BLOCKS_FEED_NAME}:
            return self.new_block_streaming
        elif feed_name in {rpc_constants.TRANSACTION_STATUS_FEED_NAME}:
            return self.transaction_state_feed
        elif feed_name in {rpc_constants.ETH_ON_BLOCK_FEED_NAME}:
            return self.on_block_feed
        elif feed_name in {rpc_constants.ETH_TRANSACTION_RECEIPTS_FEED_NAME}:
            return self.transaction_receipts_feed
        else:
            return None



from abc import ABCMeta, abstractmethod
from typing import Optional, NamedTuple, Dict, Type

from bxcommon.exceptions import ParseError, UnrecognizedCommandError
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils import convert
from bxcommon.utils.buffers.input_buffer import InputBuffer


class MessagePreview(NamedTuple):
    is_full_message: bool
    command: Optional[bytes]
    payload_length: Optional[int]


class AbstractMessageFactory:
    """
    Message factory abstract base class.

    Unpacks buffer content based on subclass implementations, pulling out the command type string and loading the
    message type class from the loader.
    """

    __metaclass__ = ABCMeta
    message_type_mapping: Dict[bytes, Type[AbstractMessage]]

    def __init__(self, message_type_mapping: Optional[Dict[bytes, Optional[Type[AbstractMessage]]]] = None) -> None:
        self.base_message_type = self.get_base_message_type()
        if message_type_mapping:
            self.message_type_mapping = {k: v for k, v in message_type_mapping.items() if v is not None}
        else:
            self.message_type_mapping = {}

    @abstractmethod
    def get_base_message_type(self) -> Type[AbstractMessage]:
        pass

    def get_message_header_preview_from_input_buffer(self, input_buffer: InputBuffer) -> MessagePreview:
        """
        Peeks at a message on the input buffer, determining if its full.
        Returns (is_full_message, command, payload_length)
        """
        if input_buffer.length < self.base_message_type.HEADER_LENGTH:
            return MessagePreview(False, None, None)
        else:
            unpacked_args = self.base_message_type.unpack(input_buffer[:self.base_message_type.HEADER_LENGTH])
            command = unpacked_args[0]
            payload_length = unpacked_args[-1]
            is_full_message = input_buffer.length >= payload_length + self.base_message_type.HEADER_LENGTH
            return MessagePreview(is_full_message, command, payload_length)

    def create_message_from_buffer(self, buf):
        """
        Parses a full message from a buffer based on its command into one of the loaded message types.
        """
        if len(buf) < self.base_message_type.HEADER_LENGTH:
            raise ParseError("Message was too short to be parsed. Raw data: {0}".format(buf))

        unpacked_args = self.base_message_type.unpack(buf)
        self.base_message_type.validate_payload(buf, unpacked_args)

        command = unpacked_args[0]
        if command not in self.message_type_mapping:
            raise UnrecognizedCommandError("Message not recognized: {0}".format(command), convert.bytes_to_hex(buf))

        return self.create_message(command, buf, unpacked_args)

    def create_message(self, command, buf, args=None):
        message_cls = self.message_type_mapping[command]
        return self.base_message_type.initialize_class(message_cls, buf, args)



import time
from collections import deque
from datetime import datetime
from typing import Tuple, Optional, Dict, NamedTuple, Set

from bxutils import logging

from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import Sha256Hash, BtcObjectHash
from bxcommon.utils.object_encoder import ObjectEncoder
from bxcommon.utils.proxy import task_pool_proxy
from bxcommon.services.extension_transaction_service import ExtensionTransactionService
from bxcommon.utils.proxy.task_queue_proxy import TaskQueueProxy
from bxcommon.utils.proxy.vector_proxy import VectorProxy
from bxcommon.utils.memory_utils import SpecialTuple
from bxcommon.utils import memory_utils

from bxgateway.messages.btc import btc_normal_message_converter
from bxgateway import btc_constants
from bxgateway.messages.btc.abstract_btc_message_converter import AbstractBtcMessageConverter, get_block_info, \
    CompactBlockCompressionResult
from bxgateway.messages.btc.compact_block_btc_message import CompactBlockBtcMessage
from bxgateway.utils.block_info import BlockInfo
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.utils.errors import message_conversion_error
from bxgateway.abstract_message_converter import BlockDecompressionResult

import task_pool_executor as tpe

logger = logging.get_logger(__name__)


class ExtensionCompactBlockRecoveryData(NamedTuple):
    tx_service: ExtensionTransactionService
    mapping_task: tpe.BtcCompactBlockMappingTask


def create_recovered_transactions() -> VectorProxy[tpe.InputBytes, memoryview]:
    encoder = ObjectEncoder(
        lambda input_bytes: memoryview(input_bytes), lambda buf: tpe.InputBytes(buf)
    )
    return VectorProxy(tpe.TransactionList(), encoder)


class BtcExtensionMessageConverter(AbstractBtcMessageConverter):

    DEFAULT_BLOCK_SIZE = btc_constants.BTC_DEFAULT_BLOCK_SIZE
    MINIMAL_SUB_TASK_TX_COUNT = btc_constants.BTC_MINIMAL_SUB_TASK_TX_COUNT

    def __init__(self, btc_magic):
        super(BtcExtensionMessageConverter, self).__init__(btc_magic)
        self._default_block_size = self.DEFAULT_BLOCK_SIZE
        self.compression_tasks = TaskQueueProxy(self._create_compression_task)
        self.compact_mapping_tasks = TaskQueueProxy(self._create_compact_mapping_task)
        self.decompression_tasks = TaskQueueProxy(self._create_decompression_task)
        self._extension_recovered_items: Dict[int, ExtensionCompactBlockRecoveryData] = {}

    def block_to_bx_block(
        self, block_msg, tx_service, enable_block_compression: bool, min_tx_age_seconds: float
    ) -> Tuple[memoryview, BlockInfo]:
        compress_start_datetime = datetime.utcnow()
        compress_start_timestamp = time.time()
        self._default_block_size = max(self._default_block_size, len(block_msg.buf))
        tsk = self.compression_tasks.borrow_task()
        tsk.init(tpe.InputBytes(block_msg.buf), tx_service.proxy, enable_block_compression, min_tx_age_seconds)
        try:
            task_pool_proxy.run_task(tsk)
        except tpe.AggregatedException as e:
            self.compression_tasks.return_task(tsk)
            raise message_conversion_error.btc_block_compression_error(block_msg.block_hash(), e)
        bx_block = tsk.bx_block()
        block = memoryview(bx_block)
        compressed_size = len(block)
        original_size = len(block_msg.rawbytes())
        block_hash = BtcObjectHash(
            binary=convert.hex_to_bytes(tsk.block_hash().hex_string())
        )

        block_info = BlockInfo(
            block_hash,
            tsk.short_ids(),
            compress_start_datetime,
            datetime.utcnow(),
            (time.time() - compress_start_timestamp) * 1000,
            tsk.txn_count(),
            tsk.compressed_block_hash().hex_string(),
            tsk.prev_block_hash().hex_string(),
            original_size,
            compressed_size,
            100 - float(compressed_size) / original_size * 100,
            tsk.ignored_short_ids()
        )
        self.compression_tasks.return_task(tsk)
        return block, block_info

    def bx_block_to_block(self, bx_block_msg, tx_service) -> BlockDecompressionResult:
        decompress_start_datetime = datetime.utcnow()
        decompress_start_timestamp = time.time()
        tsk = self.decompression_tasks.borrow_task()
        tsk.init(tpe.InputBytes(bx_block_msg), tx_service.proxy)
        try:
            task_pool_proxy.run_task(tsk)
        except tpe.AggregatedException as e:
            self.decompression_tasks.return_task(tsk)
            header_info = btc_normal_message_converter.parse_bx_block_header(bx_block_msg, deque())
            raise message_conversion_error.btc_block_decompression_error(header_info.block_hash, e)
        total_tx_count = tsk.tx_count()
        unknown_tx_hashes = [Sha256Hash(bytearray(unknown_tx_hash.binary()))
                             for unknown_tx_hash in tsk.unknown_tx_hashes()]
        unknown_tx_sids = tsk.unknown_tx_sids()
        block_hash = BtcObjectHash(
            binary=convert.hex_to_bytes(tsk.block_hash().hex_string())
        )
        if tsk.success():
            btc_block_msg = BlockBtcMessage(buf=memoryview(tsk.block_message()))
            logger.debug(
                "Successfully parsed block broadcast message. {} transactions "
                "in block {}",
                total_tx_count,
                block_hash
            )
        else:
            btc_block_msg = None
            logger.debug(
                "Block recovery needed for {}. Missing {} sids, {} tx hashes. "
                "Total txs in block: {}",
                block_hash,
                len(unknown_tx_sids),
                len(unknown_tx_hashes),
                total_tx_count
            )
        block_info = get_block_info(
            bx_block_msg,
            block_hash,
            tsk.short_ids(),
            decompress_start_datetime,
            decompress_start_timestamp,
            total_tx_count,
            btc_block_msg
        )
        self.decompression_tasks.return_task(tsk)
        return BlockDecompressionResult(btc_block_msg, block_info, unknown_tx_sids, unknown_tx_hashes)

    # pyre-fixme[14]: `compact_block_to_bx_block` overrides method defined in
    #  `AbstractBtcMessageConverter` inconsistently.
    def compact_block_to_bx_block(
        self,
        compact_block: CompactBlockBtcMessage,
        transaction_service: ExtensionTransactionService
    ) -> CompactBlockCompressionResult:
        compress_start_datetime = datetime.utcnow()
        tsk = self.compact_mapping_tasks.borrow_task()
        tsk.init(tpe.InputBytes(compact_block.buf), transaction_service.proxy, compact_block.magic())
        try:
            task_pool_proxy.run_task(tsk)
        except tpe.AggregatedException as e:
            self.compact_mapping_tasks.return_task(tsk)
            raise message_conversion_error.btc_compact_block_compression_error(compact_block.block_hash(), e)
        success = tsk.success()
        recovered_item = ExtensionCompactBlockRecoveryData(transaction_service, tsk)
        block_info = BlockInfo(
            compact_block.block_hash(),
            [],
            compress_start_datetime,
            compress_start_datetime,
            0,
            None,
            None,
            None,
            len(compact_block.rawbytes()),
            None,
            None,
            []
        )
        if success:
            result = CompactBlockCompressionResult(
                False, block_info, None, None, [], create_recovered_transactions()
            )
            return self._extension_recovered_compact_block_to_bx_block(result, recovered_item)
        else:
            recovery_index = self._last_recovery_idx
            self._extension_recovered_items[recovery_index] = recovered_item
            self._last_recovery_idx += 1
            return CompactBlockCompressionResult(
                False,
                block_info,
                None,
                recovery_index,
                tsk.missing_indices(),
                create_recovered_transactions()
            )

    def recovered_compact_block_to_bx_block(  # pyre-ignore
            self,
            failed_mapping_result: CompactBlockCompressionResult,
    ) -> CompactBlockCompressionResult:
        failed_block_info = failed_mapping_result.block_info
        block_info = BlockInfo(
            failed_block_info.block_hash,  # pyre-ignore
            failed_block_info.short_ids,  # pyre-ignore
            datetime.utcnow(),
            datetime.utcnow(),
            0,
            None,
            None,
            None,
            failed_block_info.original_size,  # pyre-ignore
            None,
            None,
            []
        )
        failed_mapping_result.block_info = block_info
        recovered_item = self._extension_recovered_items.pop(failed_mapping_result.recovery_index)  # pyre-ignore
        return self._extension_recovered_compact_block_to_bx_block(
            failed_mapping_result,
            recovered_item
        )

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return memory_utils.add_special_objects(
            self.compression_tasks, self.decompression_tasks, self.compact_mapping_tasks, ids=ids
        )

    def _extension_recovered_compact_block_to_bx_block(
        self,
        mapping_result: CompactBlockCompressionResult,
        recovery_item: ExtensionCompactBlockRecoveryData
    ) -> CompactBlockCompressionResult:
        mapping_task = recovery_item.mapping_task
        compression_task: tpe.BtcCompactBlockCompressionTask = mapping_task.compression_task()
        # pyre-fixme[16]: `List` has no attribute `vector`.
        compression_task.add_recovered_transactions(mapping_result.recovered_transactions.vector)
        mapping_block_info = mapping_result.block_info
        try:
            task_pool_proxy.run_task(compression_task)
        except tpe.AggregatedException as e:
            self.compact_mapping_tasks.return_task(mapping_task)
            # pyre-fixme[16]: `Optional` has no attribute `block_hash`.
            raise message_conversion_error.btc_compact_block_compression_error(mapping_block_info.block_hash, e)
        bx_block = memoryview(compression_task.bx_block())
        block_hash = mapping_block_info.block_hash
        txn_count = compression_task.txn_count()
        compressed_block_hash = compression_task.compressed_block_hash().hex_string()
        prev_block_hash = compression_task.prev_block_hash().hex_string()
        short_ids = compression_task.short_ids()
        compress_end_datetime = datetime.utcnow()
        # pyre-fixme[16]: `Optional` has no attribute `start_datetime`.
        compress_start_datetime = mapping_block_info.start_datetime
        # pyre-fixme[16]: `Optional` has no attribute `original_size`.
        original_size = mapping_block_info.original_size
        compressed_size = len(bx_block)
        block_info = BlockInfo(
            block_hash,
            short_ids,
            compress_start_datetime,
            compress_end_datetime,
            (compress_end_datetime - compress_start_datetime).total_seconds() * 1000,
            txn_count,
            compressed_block_hash,
            prev_block_hash,
            original_size,
            compressed_size,
            100 - float(compressed_size) / original_size * 100,
            []
        )
        self.compact_mapping_tasks.return_task(mapping_task)
        return CompactBlockCompressionResult(True, block_info, bx_block, None, [], [])

    def _create_compression_task(self) -> tpe.BtcBlockCompressionTask:
        return tpe.BtcBlockCompressionTask(self._default_block_size, self.MINIMAL_SUB_TASK_TX_COUNT)

    def _create_decompression_task(self) -> tpe.BtcBlockDecompressionTask:
        return tpe.BtcBlockDecompressionTask(
            self._default_block_size, self.MINIMAL_SUB_TASK_TX_COUNT
        )

    def _create_compact_mapping_task(self) -> tpe.BtcCompactBlockMappingTask:
        return tpe.BtcCompactBlockMappingTask(self._default_block_size)



from datetime import datetime
from threading import active_count
from typing import Type, Any, Dict, TYPE_CHECKING

from bxcommon.constants import INFO_STATS_INTERVAL_S
from bxcommon.utils.stats.statistics_service import StatisticsService, StatsIntervalData
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode


class NodeInfo(StatisticsService[StatsIntervalData, "AbstractNode"]):
    def __init__(self, interval: int = 0) -> None:
        super(NodeInfo, self).__init__(
            "NodeInfo",
            interval=interval,
            look_back=0,
            reset=False,
            stat_logger=logging.get_logger(LogRecordType.NodeInfo, __name__),
        )

    def get_interval_data_class(self) -> Type[StatsIntervalData]:
        return StatsIntervalData

    def get_info(self) -> Dict[str, Any]:
        node = self.node
        assert node is not None
        payload = dict(node.opts.__dict__)
        payload["current_time"] = datetime.utcnow()
        payload["node_peers"] = {
            connection_type: len(connections)
            for (
                connection_type,
                connections,
            ) in node.connection_pool.by_connection_type.items()
        }
        payload["threads"] = active_count()

        return payload


node_info_statistics = NodeInfo(interval=INFO_STATS_INTERVAL_S)



import time
from datetime import datetime

from bxutils import logging
from bxutils.logging import CustomLogger

logger = logging.get_logger(__name__)


def log_operation_duration(
    stats_logger: CustomLogger,
    operation_name: str,
    start_time: float,
    duration_warn_threshold: float,
    **kwargs
):
    duration = time.time() - start_time
    logger.trace("Performance: {} took {:.3f} s to execute. {}", operation_name, duration, kwargs)

    if duration > duration_warn_threshold:
        logger.debug(
            "Performance Warning: {} took over defined threshold ({} s) to execute, {:.3f}s. {}",
            operation_name,
            duration_warn_threshold,
            duration,
            kwargs,
        )

        stats_logger.statistics(
            {
                "type": "OperationDurationExceededWarningThreshold",
                "start_time": datetime.utcnow(),
                "operation": operation_name,
                "duration": duration,
                "duration_warn_threshold": duration_warn_threshold,
                "data": {k: str(v) for (k, v) in kwargs.items()},
            }
        )



from typing import Union

from cryptography import x509
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKeyWithSerialization
from cryptography.hazmat.primitives.serialization import Encoding, PrivateFormat, NoEncryption
from cryptography.x509 import Certificate, CertificateSigningRequest


def to_binary(data: Union[str, bytes]) -> bytes:
    if isinstance(data, str):
        return data.encode("utf-8")
    else:
        return data


def serialize_csr(csr: CertificateSigningRequest) -> str:
    return csr.public_bytes(Encoding.PEM).decode("utf-8")


def deserialize_csr(csr: Union[str, bytes]) -> CertificateSigningRequest:
    return x509.load_pem_x509_csr(to_binary(csr), backends.default_backend())


def serialize_cert(cert: Certificate) -> str:
    return cert.public_bytes(Encoding.PEM).decode("utf-8")


def serialize_key(key: EllipticCurvePrivateKeyWithSerialization) -> str:
    return key.private_bytes(Encoding.PEM, PrivateFormat.TraditionalOpenSSL, NoEncryption()).decode("utf-8")


def deserialize_cert(cert: Union[str, bytes]) -> Certificate:
    return x509.load_pem_x509_certificate(to_binary(cert), backends.default_backend())


def deserialize_key(key: Union[str, bytes]) -> EllipticCurvePrivateKeyWithSerialization:
    return serialization.load_pem_private_key(to_binary(key), None, backends.default_backend())



import blxr_rlp as rlp
from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxcommon.utils.blockchain_utils.eth import rlp_utils


class GetNodeDataEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.GET_NODE_DATA

    fields = [("hashes", rlp.sedes.CountableList(rlp.sedes.binary))]

    def get_hashes(self):
        return rlp_utils.get_first_list_field_items_bytes(self._memory_view)


import struct
from typing import List, Set, Union

import task_pool_executor as tpe

from bxcommon import constants
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.messages.bloxroute.txs_message import TxsMessage
from bxcommon.models.tx_validation_status import TxValidationStatus
from bxcommon.services.extension_transaction_service import ExtensionTransactionService
from bxcommon.services.transaction_service import TransactionFromBdnGatewayProcessingResult
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode
from bxgateway.connections.ont.ont_gateway_node import OntGatewayNode
from bxgateway.services.gateway_transaction_service import GatewayTransactionService, \
    ProcessTransactionMessageFromNodeResult, MissingTransactions


class ExtensionGatewayTransactionService(ExtensionTransactionService, GatewayTransactionService):

    def process_gateway_transaction_from_bdn(
        self,
        transaction_hash: Sha256Hash,
        short_id: int,
        transaction_contents: Union[bytearray, memoryview],
        is_compact: bool
    ) -> TransactionFromBdnGatewayProcessingResult:

        transaction_key = self.get_transaction_key(transaction_hash)

        ext_result = self.proxy.process_gateway_transaction_from_bdn(
            # pyre-fixme[6]: Expected `bxcommon.models.transaction_key.TransactionKey`
            transaction_key.transaction_cache_key,
            tpe.InputBytes(transaction_contents),
            short_id,
            is_compact
        )

        result = TransactionFromBdnGatewayProcessingResult(
            ext_result.get_ignore_seen(),
            ext_result.get_existing_short_id(),
            ext_result.get_assigned_short_id(),
            ext_result.get_existing_contents(),
            ext_result.get_set_contents()
        )

        if result.set_content:
            has_short_id, previous_size = ext_result.get_set_contents_result()
            self.set_transaction_contents_base_by_key(
                transaction_key,
                has_short_id,
                previous_size,
                False,
                transaction_contents,
                len(transaction_contents)
            )

        if result.assigned_short_id:
            has_contents = result.existing_contents or result.set_content
            self.assign_short_id_base_by_key(
                transaction_key,
                short_id,
                has_contents,
                False
            )

        return result

    def process_transactions_message_from_node(
        self,
        msg,
        min_tx_network_fee: int,
        enable_transaction_validation: bool
    ) -> List[ProcessTransactionMessageFromNodeResult]:
        opts = self.node.opts
        msg_bytes = msg.rawbytes()
        result = []

        # Node can return an empty list with the new version on geth (v1.10.3). in this case not need to call extensions
        if msg_bytes == b'\xc0':
            return result

        if (isinstance(self.node, OntGatewayNode) or isinstance(self.node, EthGatewayNode)) and \
                opts.process_node_txs_in_extension:
            ext_processing_results = memoryview(self.proxy.process_gateway_transaction_from_node(
                tpe.InputBytes(msg_bytes),
                min_tx_network_fee,
                enable_transaction_validation
            ))
        else:
            return GatewayTransactionService.process_transactions_message_from_node(
                self, msg, min_tx_network_fee, enable_transaction_validation
            )

        offset = 0

        txs_count, = struct.unpack_from("<I", ext_processing_results, offset)
        offset += constants.UL_INT_SIZE_IN_BYTES

        while offset < len(ext_processing_results):
            seen, = struct.unpack_from("<?", ext_processing_results, offset)
            offset += 1

            tx_hash = Sha256Hash(ext_processing_results[offset:offset + crypto.SHA256_HASH_LEN])
            offset += crypto.SHA256_HASH_LEN
            tx_offset, = struct.unpack_from("<I", ext_processing_results, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES
            tx_len, = struct.unpack_from("<I", ext_processing_results, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES
            tx_validation_status, = struct.unpack_from("<I", ext_processing_results, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES

            tx_contents = msg_bytes[tx_offset:tx_offset + tx_len]

            result.append(
                ProcessTransactionMessageFromNodeResult(
                    seen,
                    tx_hash,
                    tx_contents,
                    TxMessage(
                        message_hash=tx_hash,
                        network_num=self.network_num,
                        tx_val=tx_contents,
                        account_id=self.node.account_id
                    ),
                    TxValidationStatus(tx_validation_status)
                )
            )

            if not seen:
                transaction_key = self.get_transaction_key(tx_hash)
                self.set_transaction_contents_base_by_key(transaction_key, False, 0, False, tx_contents, tx_len)

        assert txs_count == len(result)

        return result

    def process_txs_message(
        self,
        msg: TxsMessage
    ) -> Set[MissingTransactions]:
        msg_bytes = msg.rawbytes()
        missing_transactions: Set[MissingTransactions] = set()
        result_buffer = memoryview(self.proxy.process_txs_msg(tpe.InputBytes(msg_bytes)))
        # result_buffer is a bytearray that consists of
        # number of missing transactions - 2 bytes
        # a list of:
        #   short_id - 4 bytes
        #   Sha256Hash - 32 bytes
        #   content length that was added to transaction service - 4 bytes
        offset = 0
        missing_transactions_count, = struct.unpack_from("<H", result_buffer, offset)
        offset += constants.UL_SHORT_SIZE_IN_BYTES
        if missing_transactions_count == 0:
            return missing_transactions

        while offset < len(result_buffer):
            short_id, = struct.unpack_from("<L", result_buffer, offset)
            offset += constants.SID_LEN

            transaction_hash = Sha256Hash(result_buffer[offset:offset + crypto.SHA256_HASH_LEN])
            transaction_key = self.get_transaction_key(transaction_hash)
            offset += crypto.SHA256_HASH_LEN

            content_length, = struct.unpack_from("<L", result_buffer, offset)
            offset += constants.UL_INT_SIZE_IN_BYTES
            if content_length > 0:
                self.set_transaction_contents_base_by_key(
                    transaction_key,
                    True,
                    0,
                    False,
                    None,
                    content_length
                )
            missing_transactions.add(MissingTransactions(short_id, transaction_hash))

        return missing_transactions



# pyre-ignore-all-errors
import socket
from argparse import Namespace
from typing import Optional, List

from mock import MagicMock

from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.constants import LOCALHOST
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.peer_info import ConnectionPeerInfo
from bxcommon.test_utils import helpers
from bxcommon.test_utils.mocks.mock_node_ssl_service import MockNodeSSLService
from bxcommon.utils import convert
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxgateway.services.abstract_block_cleanup_service import AbstractBlockCleanupService
from bxgateway.services.push_block_queuing_service import PushBlockQueuingService
from bxutils import logging
from bxutils.services.node_ssl_service import NodeSSLService

logger = logging.get_logger(__name__)


class NullConnection(AbstractConnection):
    pass


# noinspection PyTypeChecker
class NullGatewayNode(AbstractGatewayNode):
    """
    Test Gateway Node that doesn't connect use its blockchain or relay connection.
    """

    def __init__(self, opts: Namespace, node_ssl_service: Optional[NodeSSLService] = None):
        if node_ssl_service is None:
            node_ssl_service = MockNodeSSLService(self.NODE_TYPE, MagicMock())
        helpers.set_extensions_parallelism()
        super().__init__(opts, node_ssl_service)

    def build_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        return NullConnection

    def build_relay_connection(self, socket_connection: AbstractSocketConnectionProtocol) -> AbstractRelayConnection:
        return NullConnection

    def build_remote_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        return NullConnection

    def sync_and_send_request_for_relay_peers(self, network_num: int):
        return 0

    def build_block_queuing_service(self) -> PushBlockQueuingService:
        pass

    def build_block_cleanup_service(self) -> AbstractBlockCleanupService:
        pass

    def _send_request_for_gateway_peers(self):
        return 0

    def get_outbound_peer_info(self) -> List[ConnectionPeerInfo]:
        return [ConnectionPeerInfo(
            IpEndpoint(peer.ip, peer.port),
            convert.peer_node_to_connection_type(self.NODE_TYPE, peer.node_type)) for peer in self.outbound_peers
        ]


class NullBlockchainNode:

    def __init__(self, port):
        logger.debug("Starting null blockchain node on {}".format(port))

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.bind((LOCALHOST, port))
        self._sock.listen(50)

        self.connection = None

    def accept(self):
        self.connection, address = self._sock.accept()
        logger.debug("Null blockchain got a connection from {}".format(address))
        return address



from typing import Dict, Any, List, Union

from bxcommon.utils.object_hash import Sha256Hash
from bxgateway import log_messages
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxcommon.messages.eth.serializers.block import Block
from bxutils import logging

logger = logging.get_logger(__name__)


class EthBlockFeedEntry:
    hash: str
    header: Dict[str, Any]
    transactions: List[Any]
    uncles: List[Any]

    def __init__(
        self, hash: Sha256Hash, block: Union[Block, InternalEthBlockInfo, memoryview]
    ) -> None:
        self.hash = f"0x{str(hash)}"
        try:
            if isinstance(block, Block):
                block_info = block
            elif isinstance(block, InternalEthBlockInfo):
                block_info = block.to_new_block_msg().get_block()
            else:
                block_info = InternalEthBlockInfo(block).to_new_block_msg().get_block()
            block_json = block_info.to_json()
            self.header = block_json["header"]
            self.transactions = block_json["transactions"]
            self.uncles = block_json["uncles"]
        except Exception as e:
            block_str = block
            if isinstance(block, memoryview):
                block_str = block.tobytes()

            logger.error(
                log_messages.COULD_NOT_DESERIALIZE_BLOCK,
                hash,
                block_str,
                e
            )
            raise e

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, EthBlockFeedEntry)
            and other.hash == self.hash
            and other.header == self.header
            and other.transactions == self.transactions
            and other.uncles == self.uncles
        )



from dataclasses import dataclass
from typing import Optional

from bxgateway.utils.logging.status.connection_info import ConnectionInfo


@dataclass
class RelayConnection(ConnectionInfo):
    peer_id: Optional[str]

    def __init__(self, ip_address: str, port: str, fileno: Optional[str] = None,
                 peer_id: Optional[str] = None, connection_time: Optional[str] = None):
        self.ip_address = ip_address
        self.port = port
        self.fileno = fileno
        self.peer_id = peer_id
        self.connection_time = connection_time

    def __eq__(self, other: "RelayConnection") -> bool:
        return self.ip_address == other.ip_address \
               and self.port == other.port \
               and self.peer_id == other.peer_id



import struct
from abc import ABCMeta

from bxcommon import constants
from bxcommon.constants import MSG_NULL_BYTE
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage


class AbstractBlockchainSyncMessage(AbstractBloxrouteMessage):
    """
    Message type for requesting/receiving direct blockchain messages for syncing chainstate.
    """

    __metaclass__ = ABCMeta

    MESSAGE_TYPE = ""
    BASE_LENGTH = 12

    def __init__(self, command=None, payload=None, buf=None):
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + constants.MSG_TYPE_LEN + len(payload))

            off = self.HEADER_LENGTH
            struct.pack_into("<12s", buf, off, command)

            off += constants.MSG_TYPE_LEN
            buf[off:off + len(payload)] = payload

        self.buf = buf
        self._command = None
        self._payload = None
        payload_length = len(buf) - self.HEADER_LENGTH
        super(AbstractBlockchainSyncMessage, self).__init__(self.MESSAGE_TYPE, payload_length, self.buf)

    def command(self):
        """
        Blockchain command. Can be either a string or an int.
        :return:
        """
        if self._command is None:
            off = self.HEADER_LENGTH
            self._command, = struct.unpack_from("<12s", self.buf, off)
            self._command = str(self._command).rstrip(MSG_NULL_BYTE)

            try:
                self._command = int(self._command)
            except:
                pass

        return self._command

    def payload(self):
        if self._payload is None:
            off = self.HEADER_LENGTH + constants.MSG_TYPE_LEN
            self._payload = self._memoryview[off:off + self.payload_len()]
        return self._payload



import hashlib

import bitcoin
import pyelliptic

from bxgateway.eth_exceptions import DecryptionError
from bxcommon.utils.blockchain_utils.eth import crypto_utils, rlp_utils, eth_common_constants


class ECCx(pyelliptic.ECC):
    """
    Extensions for ECC class from pyelliptic library
    Modified to work with raw_pubkey format used in RLPx
    and binding default curve and cipher
    """

    def __init__(self, raw_public_key=None, raw_private_key=None):

        self._init_ecc(raw_public_key, raw_private_key)

    def get_raw_private_key(self):
        """
        Raw private key used by instance of ECCx
        :return: private key
        """

        if self.privkey:
            return rlp_utils.str_to_bytes(self.privkey)
        return self.privkey

    def get_raw_public_key(self):
        """
        Raw public ley user by instance of ECCx
        :return: public key
        """

        if self.pubkey_x and self.pubkey_y:
            return rlp_utils.str_to_bytes(self.pubkey_x + self.pubkey_y)
        return self.pubkey_x + self.pubkey_y

    def get_ecdh_key(self, raw_public_key):
        """
        Compute public key with the local private key
        and returns a 256bits shared ECDH (Elliptic-Curve Diffie-Hellman) key
        :param raw_public_key: other parties raw public key
        :return shared key
        """

        if len(raw_public_key) != eth_common_constants.PUBLIC_KEY_LEN:
            raise ValueError("Public key of len {0} is expected but was {1}"
                             .format(eth_common_constants.PUBLIC_KEY_LEN, len(raw_public_key)))

        _, pubkey_x, pubkey_y, _ = self._decode_pubkey(raw_public_key)
        key = self.raw_get_ecdh_key(pubkey_x, pubkey_y)
        assert len(key) == eth_common_constants.SHARED_KEY_LEN
        return key

    def is_valid_key(self, raw_public_key, raw_private_key=None):
        """
        Validates raw public key
        :param raw_public_key: raw public key
        :param raw_private_key: raw private key (optional)
        :return: True is valid, False otherwise
        """

        if len(raw_public_key) != eth_common_constants.PUBLIC_KEY_LEN:
            raise ValueError("Public key of len {0} is expected but was {1}"
                             .format(eth_common_constants.PUBLIC_KEY_LEN, len(raw_public_key)))

        try:
            assert len(raw_public_key) == eth_common_constants.PUBLIC_KEY_LEN
            pubkey_x = raw_public_key[:eth_common_constants.PUBLIC_KEY_X_Y_LEN]
            pubkey_y = raw_public_key[eth_common_constants.PUBLIC_KEY_X_Y_LEN:]
            failed = bool(self.raw_check_key(raw_private_key, pubkey_x, pubkey_y))
        except (AssertionError, Exception):
            failed = True
        return not failed

    def encrypt(cls, data, raw_public_key, shared_mac_data=""):
        """
        ECIES Encrypt, where P = recipient public key is:
        1) generate r = random value
        2) generate shared-secret = kdf( ecdhAgree(r, P) )
        3) generate R = rG [same op as generating a public key]
        4) send 0x04 || R || AsymmetricEncrypt(shared-secret, plaintext) || tag


        currently used by go:
        ECIES_AES128_SHA256 = &ECIESParams{
            Hash: sha256.New,
            hashAlgo: crypto.SHA256,
            Cipher: aes.NewCipher,
            BlockSize: aes.BlockSize,
            KeyLen: 16,
            }

        :param data: data to encrypt
        :param raw_public_key: public key used of encryption
        :param shared_mac_data: shared mac
        :return: encrypted data
        """

        if not data:
            raise ValueError("Data is required")

        if len(raw_public_key) != eth_common_constants.PUBLIC_KEY_LEN:
            raise ValueError("Public key of len {0} is expected but was {1}"
                             .format(eth_common_constants.PUBLIC_KEY_LEN, len(raw_public_key)))

        # 1) generate r = random value
        ephem = ECCx()

        # 2) generate shared-secret = kdf( ecdhAgree(r, P) )
        pubkey_x = raw_public_key[:eth_common_constants.PUBLIC_KEY_X_Y_LEN]
        pubkey_y = raw_public_key[eth_common_constants.PUBLIC_KEY_X_Y_LEN:]
        key_material = ephem.raw_get_ecdh_key(pubkey_x=pubkey_x, pubkey_y=pubkey_y)

        key = crypto_utils.ecies_kdf(key_material, eth_common_constants.SHARED_KEY_LEN)
        assert len(key) == eth_common_constants.SHARED_KEY_LEN
        key_enc, key_mac = key[:eth_common_constants.ENC_KEY_LEN], key[eth_common_constants.ENC_KEY_LEN:]

        key_mac = hashlib.sha256(key_mac).digest()

        # 3) generate R = rG [same op as generating a public key]
        ephem_pubkey = ephem.get_raw_public_key()

        # encrypt
        iv = pyelliptic.Cipher.gen_IV(eth_common_constants.ECIES_CIPHER_NAME)
        assert len(iv) == eth_common_constants.IV_LEN

        ctx = pyelliptic.Cipher(key_enc, iv, eth_common_constants.CIPHER_ENCRYPT_DO, eth_common_constants.ECIES_CIPHER_NAME)
        cipher_text = ctx.ciphering(data)
        assert len(cipher_text) == len(data)

        # 4) send 0x04 || R || AsymmetricEncrypt(shared-secret, plaintext) || tag
        msg = rlp_utils.ascii_chr(eth_common_constants.ECIES_HEADER) + ephem_pubkey + iv + cipher_text

        # the MAC of a message (called the tag) as per SEC 1, 3.5.
        tag = pyelliptic.hmac_sha256(key_mac, msg[eth_common_constants.ECIES_HEADER_LEN + eth_common_constants.PUBLIC_KEY_LEN:] +
                                     rlp_utils.str_to_bytes(shared_mac_data))
        assert len(tag) == eth_common_constants.MAC_LEN
        msg += tag

        assert len(msg) - eth_common_constants.ECIES_ENCRYPT_OVERHEAD_LENGTH == len(data)
        return msg

    def decrypt(self, data, shared_mac_data=b""):
        """
        Decrypt data with ECIES method using the local private key

        ECIES Decrypt (performed by recipient):
        1) generate shared-secret = kdf( ecdhAgree(myPrivKey, msg[1:65]) )
        2) verify tag
        3) decrypt

        ecdhAgree(r, recipientPublic) == ecdhAgree(recipientPrivate, R)
        [where R = r*G, and recipientPublic = recipientPrivate*G]

        :param data: data to decrypt
        :param shared_mac_data: shared mac
        :return: decrypted data
        """

        if not data:
            raise ValueError("Data is required")

        if data[:eth_common_constants.ECIES_HEADER_LEN] != eth_common_constants.ECIES_HEADER_BYTE:
            raise DecryptionError("Wrong ECIES header")

        #  1) generate shared-secret = kdf( ecdhAgree(myPrivKey, msg[1:65]) )
        shared = data[eth_common_constants.ECIES_HEADER_LEN:eth_common_constants.ECIES_HEADER_LEN + eth_common_constants.PUBLIC_KEY_LEN]

        shared_pubkey_x = shared[:eth_common_constants.PUBLIC_KEY_X_Y_LEN]
        shared_pubkey_y = shared[eth_common_constants.PUBLIC_KEY_X_Y_LEN:]

        key_material = self.raw_get_ecdh_key(pubkey_x=shared_pubkey_x, pubkey_y=shared_pubkey_y)
        assert len(key_material) == eth_common_constants.KEY_MATERIAL_LEN
        key = crypto_utils.ecies_kdf(key_material, eth_common_constants.SHARED_KEY_LEN)
        assert len(key) == eth_common_constants.SHARED_KEY_LEN
        key_enc, key_mac = key[:eth_common_constants.ENC_KEY_LEN], key[eth_common_constants.ENC_KEY_LEN:]

        key_mac = hashlib.sha256(key_mac).digest()
        assert len(key_mac) == eth_common_constants.MAC_LEN

        tag = data[-eth_common_constants.MAC_LEN:]
        assert len(tag) == eth_common_constants.MAC_LEN

        # 2) verify tag
        header_len = eth_common_constants.ECIES_HEADER_LEN + eth_common_constants.PUBLIC_KEY_LEN

        mac_data = data[header_len:- eth_common_constants.MAC_LEN] + \
                   shared_mac_data
        if not pyelliptic.equals(pyelliptic.hmac_sha256(key_mac, mac_data), tag):
            raise DecryptionError("Fail to verify data")

        # 3) decrypt
        block_size = pyelliptic.OpenSSL.get_cipher(eth_common_constants.ECIES_CIPHER_NAME).get_blocksize()

        iv = data[header_len:header_len + block_size]
        cipher_text = data[header_len + block_size:- eth_common_constants.MAC_LEN]
        assert eth_common_constants.ECIES_HEADER_LEN + len(shared) + len(iv) + len(cipher_text) + len(tag) == len(data)
        ctx = pyelliptic.Cipher(key_enc, iv, eth_common_constants.CIPHER_DECRYPT_DO, eth_common_constants.ECIES_CIPHER_NAME)
        return ctx.ciphering(cipher_text)

    def sign(self, data):
        """
        Calculates signature for data
        :param data: data to sign
        :return: signature
        """

        if not data:
            raise ValueError("Data is required")

        signature = crypto_utils.sign(data, self.get_raw_private_key())
        assert len(signature) == eth_common_constants.SIGNATURE_LEN
        return signature

    def verify(self, signature, message):
        """
        Verifies signature for message
        :param signature: signature
        :param message: message
        :return: True if signature is valid, False otherwise
        """

        if len(signature) != eth_common_constants.SIGNATURE_LEN:
            raise ValueError("Signature len of {0} is expected but was {1}"
                             .format(eth_common_constants.SIGNATURE_LEN, len(signature)))

        return crypto_utils.verify_signature(self.get_raw_public_key(), signature, message)

    def _init_ecc(self, raw_public_key, raw_private_key):
        if raw_private_key:
            assert not raw_public_key
            raw_public_key = crypto_utils.private_to_public_key(raw_private_key)
        if raw_public_key:
            assert len(raw_public_key) == eth_common_constants.PUBLIC_KEY_LEN
            _, pubkey_x, pubkey_y, _ = self._decode_pubkey(raw_public_key)
        else:
            pubkey_x, pubkey_y = None, None

        while True:
            pyelliptic.ECC.__init__(self, pubkey_x=pubkey_x, pubkey_y=pubkey_y,
                                    raw_privkey=raw_private_key, curve=eth_common_constants.ECIES_CURVE)

            # when raw_private_key is generated by pyelliptic it sometimes has 31 bytes so we try again!
            if self.get_raw_private_key() and len(self.get_raw_private_key()) != eth_common_constants.PRIVATE_KEY_LEN:
                continue
            try:
                if self.get_raw_private_key():
                    bitcoin.get_privkey_format(self.get_raw_private_key())  # failed for some keys
                valid_private_key = True
            except AssertionError:
                valid_private_key = False
            if len(self.get_raw_public_key()) == eth_common_constants.PUBLIC_KEY_LEN and valid_private_key:
                break
            elif raw_private_key or raw_public_key:
                raise Exception("invalid private or public key")

        assert len(self.get_raw_public_key()) == eth_common_constants.PUBLIC_KEY_LEN

    def _decode_pubkey(cls, raw_pubkey):
        pubkey_x = raw_pubkey[:eth_common_constants.PUBLIC_KEY_X_Y_LEN]
        pubkey_y = raw_pubkey[eth_common_constants.PUBLIC_KEY_X_Y_LEN:]
        return eth_common_constants.ECIES_CURVE, pubkey_x, pubkey_y, eth_common_constants.PUBLIC_KEY_LEN



from enum import Flag


class StatEventLogicFlags(Flag):
    NONE = 0
    BLOCK_INFO = 1
    MATCH = 2
    SUMMARY = 4
    PROPAGATION_START = 8
    PROPAGATION_END = 16

    def __str__(self) -> str:
        return str(self.value)



from bxcommon.services import sdn_http_service
from bxcommon.common_opts import CommonOpts
from bxcommon import common_init_tasks
from bxgateway.gateway_opts import GatewayOpts

from bxutils.services.node_ssl_service import NodeSSLService


def set_account_info(opts: CommonOpts, node_ssl_service: NodeSSLService) -> None:
    account_id = node_ssl_service.get_account_id()
    if account_id:
        # TODO: use local cache for account_model
        account_model = sdn_http_service.fetch_account_model(account_id)
        if account_model:
            opts.set_account_options(account_model)


def validate_network_opts(opts: CommonOpts, _node_ssl_service: NodeSSLService) -> None:
    opts.validate_network_opts()


def set_gateway_info(opts: GatewayOpts, _node_ssl_service: NodeSSLService) -> None:
    node_config = sdn_http_service.fetch_gateway_settings(opts.node_id)
    opts.min_peer_relays_count = node_config.min_peer_relays_count


init_tasks = [
    set_account_info,
    validate_network_opts,
    common_init_tasks.set_network_info,
    common_init_tasks.set_node_model,
    set_gateway_info,
]



import random
from typing import TYPE_CHECKING, cast, Optional

from bxcommon import constants
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.connections.internal_node_connection import InternalNodeConnection
from bxcommon.feed.feed import FeedKey
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.messages.bloxroute.ack_message import AckMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.services import sdn_http_service
from bxcommon.utils import crypto
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.feed.feed_source import FeedSource
from bxgateway import gateway_constants
from bxgateway import log_messages
from bxcommon.feed.eth.eth_pending_transaction_feed import EthPendingTransactionFeed
from bxcommon.feed.eth.eth_raw_transaction import EthRawTransaction
from bxgateway.messages.gateway.confirmed_block_message import ConfirmedBlockMessage
from bxgateway.messages.gateway.confirmed_tx_message import ConfirmedTxMessage
from bxgateway.messages.gateway.gateway_hello_message import GatewayHelloMessage
from bxgateway.messages.gateway.gateway_message_factory import gateway_message_factory
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType
from bxgateway.messages.gateway.gateway_version_manager import gateway_version_manager
from bxgateway.messages.gateway.request_tx_stream_message import RequestTxStreamMessage
from bxgateway.utils.stats.transaction_feed_stats_service import transaction_feed_stats_service

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class GatewayConnection(InternalNodeConnection["AbstractGatewayNode"]):
    """
    Connection handler for Gateway-Gateway connections.
    """

    CONNECTION_TYPE = ConnectionType.EXTERNAL_GATEWAY

    NULL_ORDERING = -1
    ACK_MESSAGE = AckMessage()

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "AbstractGatewayNode"):
        super(GatewayConnection, self).__init__(sock, node)

        self.hello_messages = gateway_constants.GATEWAY_HELLO_MESSAGES
        self.header_size = constants.STARTING_SEQUENCE_BYTES_LEN + constants.BX_HDR_COMMON_OFF
        self.message_handlers = {
            GatewayMessageType.HELLO: self.msg_hello,
            BloxrouteMessageType.ACK: self.msg_ack,
            GatewayMessageType.BLOCK_RECEIVED: self.msg_block_received,
            GatewayMessageType.BLOCK_PROPAGATION_REQUEST: self.msg_block_propagation_request,
            BloxrouteMessageType.BLOCK_HOLDING: self.msg_block_holding,
            BloxrouteMessageType.KEY: self.msg_key,
            GatewayMessageType.CONFIRMED_TX: self.msg_confirmed_tx,
            GatewayMessageType.CONFIRMED_BLOCK: self.msg_confirmed_block,
            GatewayMessageType.REQUEST_TX_STREAM: self.msg_request_tx_stream,
        }
        self.version_manager = gateway_version_manager
        self.protocol_version = self.version_manager.CURRENT_PROTOCOL_VERSION
        self.stream_confirmation_messages = False

        if self.from_me:
            self._initialize_ordered_handshake()
        else:
            self.ordering = self.NULL_ORDERING

    def connection_message_factory(self) -> AbstractMessageFactory:
        return gateway_message_factory

    def on_connection_established(self):
        super().on_connection_established()

        peer_model = self.peer_model
        if (
            self.node.opts.request_remote_transaction_streaming
            and peer_model is not None
            and peer_model.is_transaction_streamer()
        ):
            self.enqueue_msg(RequestTxStreamMessage())

    def msg_hello(self, msg):
        """
        Updates connection pool references to this connection to match the port described in the hello message.
        Identifies any duplicate connections (i.e. from if both nodes connect to each other) and drops them
        based on the ordering of the messages.

        Connections may end up with two entries in the connection pool if they are initiated by the remote peer, i.e.
        one on the randomly assigned incoming socket and one on the actual provided address of the peer here.
        """
        network_num = msg.network_num()
        if network_num != self.node.network_num:
            self.log_error(log_messages.NETWORK_NUMBER_MISMATCH, self.network_num, network_num)
            self.mark_for_close(should_retry=False)
            return

        ip = msg.ip()
        if ip != self.peer_ip:
            self.log_warning(log_messages.MISMATCH_IP_IN_HELLO_MSG, ip, self.peer_ip)

        port = msg.port()
        ordering = msg.ordering()
        peer_id = msg.node_id()
        if not self._is_authenticated:
            # naively set the the peer id to what reported
            if peer_id is None:
                self.log_warning(log_messages.HELLO_MSG_WITH_NO_PEER_ID)
            self.peer_id = peer_id

        if not self.from_me:
            self.log_trace("Connection established with peer: {}.", peer_id)
            self.node.requester.send_threaded_request(
                sdn_http_service.submit_gateway_inbound_connection,
                self.node.opts.node_id,
                self.peer_id
            )

        if self.node.connection_exists(ip, port):
            connection = self.node.connection_pool.get_by_ipport(ip, port)

            # connection already has correct ip / port info assigned
            if connection == self:
                self.log_debug("Duplicate hello message received. Ignoring.")
                self.enqueue_msg(self.ack_message)
                return

            if connection.is_active():
                self.log_debug("Duplicate established connection. Dropping.")
                self.mark_for_close(should_retry=False)
                return

            # ordering numbers were the same; take over the slot and try again
            if connection.ordering == ordering:
                self.log_debug("Duplicate connection orderings could not be resolved. Investigate if this "
                               "message appears repeatedly.")
                self.node.connection_pool.update_port(self.peer_port, port, self)
                self._initialize_ordered_handshake()
                return

            # part one of duplicate connection resolution
            # if connection has same ordering or no ordering, it hasn't yet received its connection
            # so don't do anything and wait for the other connection to receive a HELLO message and itself or this one
            if connection.ordering == self.NULL_ORDERING:
                self.log_debug(
                    "Duplicate connections. Two connections have been opened on ip port: {}:{}. Awaiting other "
                    "connection's resolution.", ip, port)
                self.ordering = ordering
                return

            if connection.ordering > ordering:
                self.log_warning(log_messages.REDUNDANT_CONNECTION, self.file_no, connection.file_no)
                self.mark_for_close(should_retry=False)
                connection.on_connection_established()
                connection.enqueue_msg(connection.ack_message)
            else:
                self.log_warning(log_messages.REDUNDANT_CONNECTION, connection.file_no, self.file_no)
                connection.mark_for_close(should_retry=False)
                self.on_connection_established()
                self.enqueue_msg(self.ack_message)
                self.node.connection_pool.update_port(self.peer_port, port, self)
        else:
            self.log_debug("Connection is only one of its kind. Updating port reference.")
            self.node.connection_pool.update_port(self.peer_port, port, self)
            self.enqueue_msg(self.ack_message)
            self.on_connection_established()

        self._update_port_info(port)
        self.ordering = ordering

    def msg_ack(self, _msg):
        """
        Acks hello message and establishes connection.
        """
        self.on_connection_established()

    def msg_block_received(self, msg):
        self.node.neutrality_service.record_block_receipt(msg.block_hash(), self)

    def msg_block_propagation_request(self, msg):
        bx_block = msg.blob()
        bx_block_hash = crypto.double_sha256(bx_block)
        block_stats.add_block_event_by_block_hash(
            bx_block_hash,
            BlockStatEventType.BX_BLOCK_PROPAGATION_REQUESTED_BY_PEER,
            network_num=self.network_num,
            peers=[self],
        )
        self.node.neutrality_service.propagate_block_to_network(bx_block, self, from_peer=True)

    def msg_block_holding(self, msg):
        block_hash = msg.block_hash()
        self.node.block_processing_service.place_hold(block_hash, self)

    def msg_key(self, msg):
        """
        Handles key message receive from bloXroute.
        Looks for the encrypted block and decrypts; otherwise stores for later.
        """
        self.node.block_processing_service.process_block_key(msg, self)

    def msg_confirmed_tx(self, msg: ConfirmedTxMessage) -> None:
        tx_hash = msg.tx_hash()
        transaction_key = self.node.get_tx_service().get_transaction_key(tx_hash)
        tx_contents = msg.tx_val()

        # shouldn't ever happen, but just in case
        if tx_contents == ConfirmedTxMessage.EMPTY_TX_VAL:
            tx_contents = cast(
                Optional[memoryview],
                self.node.get_tx_service().get_transaction_by_key(transaction_key)
            )
            if tx_contents is None:
                transaction_feed_stats_service.log_pending_transaction_missing_contents()
                return

        self.node.feed_manager.publish_to_feed(
            FeedKey(EthPendingTransactionFeed.NAME, self.network_num),
            EthRawTransaction(
                tx_hash, tx_contents, FeedSource.BDN_SOCKET, local_region=True
            )
        )

        transaction_feed_stats_service.log_pending_transaction_from_internal(tx_hash)

    def msg_request_tx_stream(self, msg: RequestTxStreamMessage) -> None:
        pass

    def msg_confirmed_block(self, msg: ConfirmedBlockMessage) -> None:
        pass

    def _initialize_ordered_handshake(self):
        self.ordering = random.getrandbits(constants.UL_INT_SIZE_IN_BYTES * 8)
        self.enqueue_msg(GatewayHelloMessage(self.protocol_version, self.network_num, self.node.opts.external_ip,
                                             self.node.opts.external_port, self.ordering, self.node.opts.node_id))

    def _update_port_info(self, new_port):
        self.peer_port = new_port
        self.peer_desc = "%s %d" % (self.peer_ip, self.peer_port)

    def dispose(self):
        if not self.from_me:
            if self.peer_id:
                self.node.requester.send_threaded_request(
                    sdn_http_service.delete_gateway_inbound_connection,
                    self.node.opts.node_id,
                    self.peer_id
                )
        super().dispose()



from typing import TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class GatewayPeersRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):
    help = {
        "params": "",
        "description": "return gateway connected peers"
    }

    def validate_params(self) -> None:
        pass

    async def process_request(self) -> JsonRpcResponse:
        data = []
        connection_pool = self.node.connection_pool
        connections = (
            list(connection_pool.get_by_connection_types((ConnectionType.BLOCKCHAIN_NODE,)))
            + list(connection_pool.get_by_connection_types((ConnectionType.RELAY_ALL,)))
            + list(connection_pool.get_by_connection_types((ConnectionType.REMOTE_BLOCKCHAIN_NODE,)))
            + list(connection_pool.get_by_connection_types((ConnectionType.GATEWAY,)))
        )
        for conn in connections:
            connection_type = (
                ConnectionType.GATEWAY
                if conn.CONNECTION_TYPE in ConnectionType.GATEWAY
                else conn.CONNECTION_TYPE
            )
            data.append(
                {
                    "id": conn.peer_id,
                    "type": str(connection_type),
                    "addr": repr(conn.endpoint),
                    "direction": str(conn.direction),
                    "state": str(conn.state),
                }
            )

        return self.ok(data)



from typing import TYPE_CHECKING, Dict, Union, Optional, Any

from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.rpc_errors import RpcInvalidParams, RpcAccountIdError
from bxutils import logging

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)

TAG_TYPE = Union[str, int]


class GatewayBlxrCallRpcRequest(AbstractRpcRequest["EthGatewayNode"]):
    help = {
        "params": f"[{rpc_constants.TRANSACTION_JSON_PARAMS_KEY}: [transaction payload as json]]. "
                  f"[{rpc_constants.TAG_PARAMS_KEY}: [block number, or the string latest, earliest or pending]]",
        "description": "execute eth call through the BDN"
    }

    def validate_params(self) -> None:
        params = self.params
        if params is None or not isinstance(params, dict):
            raise RpcInvalidParams(
                self.request_id,
                "Params request field is either missing or not a dictionary type."
            )
        transaction_json = params.get(rpc_constants.TRANSACTION_JSON_PARAMS_KEY)
        if not transaction_json or not isinstance(transaction_json, dict):
            raise RpcInvalidParams(
                self.request_id,
                f"Invalid transaction request key: {rpc_constants.TRANSACTION_JSON_PARAMS_KEY} is required"
            )
        if rpc_constants.TAG_PARAMS_KEY in params:
            tag = params[rpc_constants.TAG_PARAMS_KEY]
            if isinstance(tag, int):
                pass
            elif isinstance(tag, str) and tag in {"latest", "pending", "earliest"}:
                pass
            else:
                raise RpcInvalidParams(
                    self.request_id,
                    f"Invalid value for {rpc_constants.TAG_PARAMS_KEY}: {tag}"
                )

    async def process_request(self) -> JsonRpcResponse:
        params = self.params
        assert isinstance(params, dict)
        account_id = self.get_account_id()
        if self.node.opts.blockchain_protocol != BlockchainProtocol.ETHEREUM:
            raise RpcInvalidParams(
                self.request_id,
                f"Gateway does not support {BlockchainProtocol.ETHEREUM} protocol methods"
            )
        if not account_id:
            raise RpcAccountIdError(
                self.request_id,
                "Gateway does not have an associated account. Please register the gateway with an account to submit "
                "calls through RPC."
            )

        # Hook, to verify EthNode WS connection health
        self.node.on_new_subscriber_request()

        transaction_obj: Dict[str, Any] = params[rpc_constants.TRANSACTION_JSON_PARAMS_KEY]
        tag: TAG_TYPE = params.get(rpc_constants.TAG_PARAMS_KEY, "latest")
        return await self.process_eth_call(transaction_obj, tag)

    async def process_eth_call(self, transaction_obj: Dict, tag: TAG_TYPE) -> JsonRpcResponse:
        response = await self.node.eth_ws_proxy_publisher.call_rpc(
            "eth_call",
            [transaction_obj, tag]
        )
        return self.ok({"result": response.result})

    def get_account_id(self) -> Optional[str]:
        return self.node.account_id
